@@ -49,11 +49,8 @@
 #define DEFAULT_STACK_SIZE 65536
 /* #define DEBUG_STACK_SIZE */
 
-/* I/O buffer size */
-#define BUFFSIZE 16384
-
-/* IP address and TCP port textual representation length */
-#define IPLEN 128
+/* I/O buffer size - 18432 is the maximum size of SSL record payload */
+#define BUFFSIZE 18432
 
 /* how many bytes of random input to read from files for PRNG */
 /* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
@@ -64,25 +61,70 @@
 
 /**************************************** platform */
 
-#ifdef USE_WIN32
-#define USE_IPv6
-/* #define USE_FIPS */
+#ifdef _WIN32
+#define USE_WIN32
 #endif
 
 #ifdef _WIN32_WCE
 #define USE_WIN32
 typedef int socklen_t;
-#define EINTR WSAEINTR
-#define EMFILE WSAEMFILE
 #endif
 
 #ifdef USE_WIN32
-#define HAVE_OPENSSL
+#define USE_IPv6
+/* #define USE_FIPS */
+#define _CRT_SECURE_NO_DEPRECATE
+#define _CRT_NONSTDC_NO_DEPRECATE
 #define HAVE_OSSL_ENGINE_H
+#define HAVE_OSSL_OCSP_H
 /* prevent including wincrypt.h, as it defines it's own OCSP_RESPONSE */
 #define __WINCRYPT_H__
 #endif
 
+#ifdef USE_WIN32
+#define S_EADDRINUSE  WSAEADDRINUSE
+/* winsock does not define WSAEAGAIN */
+/* in most (but not all!) BSD implementations EAGAIN==EWOULDBLOCK */
+#define S_EAGAIN        WSAEWOULDBLOCK
+#define S_ECONNRESET    WSAECONNRESET
+#define S_EINPROGRESS   WSAEINPROGRESS
+#define S_EINTR         WSAEINTR
+#define S_EINVAL        WSAEINVAL
+#define S_EISCONN       WSAEISCONN
+#define S_EMFILE        WSAEMFILE
+/* winsock does not define WSAENFILE */
+#define S_ENOBUFS       WSAENOBUFS
+/* winsock does not define WSAENOMEM */
+#define S_ENOPROTOOPT   WSAENOPROTOOPT
+#define S_ENOTSOCK      WSAENOTSOCK
+#define S_EOPNOTSUPP    WSAEOPNOTSUPP
+#define S_EWOULDBLOCK   WSAEWOULDBLOCK
+#define S_ECONNABORTED  WSAECONNABORTED
+#else /* USE_WIN32 */
+#define S_EADDRINUSE    EADDRINUSE
+#define S_EAGAIN        EAGAIN
+#define S_ECONNRESET    ECONNRESET
+#define S_EINPROGRESS   EINPROGRESS
+#define S_EINTR         EINTR
+#define S_EINVAL        EINVAL
+#define S_EISCONN       EISCONN
+#define S_EMFILE        EMFILE
+#ifdef ENFILE
+#define S_ENFILE        ENFILE
+#endif
+#ifdef ENOBUFS
+#define S_ENOBUFS       ENOBUFS
+#endif
+#ifdef ENOMEM
+#define S_ENOMEM        ENOMEM
+#endif
+#define S_ENOPROTOOPT   ENOPROTOOPT
+#define S_ENOTSOCK      ENOTSOCK
+#define S_EOPNOTSUPP    EOPNOTSUPP
+#define S_EWOULDBLOCK   EWOULDBLOCK
+#define S_ECONNABORTED  ECONNABORTED
+#endif /* USE_WIN32 */
+
 /**************************************** generic headers */
 
 #ifdef __vms
@@ -105,6 +147,7 @@ typedef int socklen_t;
 #define THREADS
 #endif
 #ifndef _REENTRANT
+/* _REENTRANT is required for thread-safe errno on Solaris */
 #define _REENTRANT
 #endif
 #ifndef _THREAD_SAFE
@@ -114,8 +157,8 @@ typedef int socklen_t;
 #endif
 
 /* TCP wrapper */
-#if HAVE_TCPD_H && HAVE_LIBWRAP
-#define USE_LIBWRAP
+#if defined HAVE_TCPD_H && defined HAVE_LIBWRAP
+#define USE_LIBWRAP 1
 #endif
 
 /* must be included before sys/stat.h for Ultrix */
@@ -143,17 +186,19 @@ typedef unsigned short u16;
 typedef unsigned long u32;
 
 #define HAVE_SNPRINTF
-#define snprintf _snprintf
+#define snprintf                    _snprintf
 #define HAVE_VSNPRINTF
-#define vsnprintf _vsnprintf
-#define strcasecmp _stricmp
-#define strncasecmp _strnicmp
-#define sleep(c) Sleep(1000*(c))
+#define vsnprintf                   _vsnprintf
+#define strcasecmp                  _stricmp
+#define strncasecmp                 _strnicmp
+#define sleep(c)                    Sleep(1000*(c))
 
-#define get_last_socket_error() WSAGetLastError()
-#define get_last_error()        GetLastError()
-#define readsocket(s,b,n)       recv((s),(b),(n),0)
-#define writesocket(s,b,n)      send((s),(b),(n),0)
+#define get_last_socket_error()     WSAGetLastError()
+#define set_last_socket_error(e)    WSASetLastError(e)
+#define get_last_error()            GetLastError()
+#define set_last_error(e)           SetLastError(e)
+#define readsocket(s,b,n)           recv((s),(b),(n),0)
+#define writesocket(s,b,n)          send((s),(b),(n),0)
 
 /* #define FD_SETSIZE 4096 */
 /* #define Win32_Winsock */
@@ -162,32 +207,14 @@ typedef unsigned long u32;
 /* Winsock2 header for IPv6 definitions */
 #include <winsock2.h>
 #include <ws2tcpip.h>
+
 #include <windows.h>
 
-/* needed for winsock2 (not for winsock) */
-#define ENOTSOCK WSAENOTSOCK
-#define ENOPROTOOPT WSAENOPROTOOPT
-#define EINPROGRESS WSAEINPROGRESS
-#define EWOULDBLOCK WSAEWOULDBLOCK
-#define EADDRINUSE WSAEADDRINUSE
-
-#define ECONNRESET WSAECONNRESET
-#define ENOTSOCK WSAENOTSOCK
-#define ENOPROTOOPT WSAENOPROTOOPT
-#define EINPROGRESS WSAEINPROGRESS
-#define EWOULDBLOCK WSAEWOULDBLOCK
-#define EISCONN WSAEISCONN
-#define EADDRINUSE WSAEADDRINUSE
-#define EOPNOTSUPP WSAEOPNOTSUPP
-
-#ifdef EINVAL
-#undef EINVAL
-#endif
-#define EINVAL WSAEINVAL
-
 #include <process.h>     /* _beginthread */
 #include <tchar.h>
 
+#include "resources.h"
+
 /**************************************** non-WIN32 headers */
 
 #else /* USE_WIN32 */
@@ -209,29 +236,33 @@ typedef unsigned long u32;
 #endif
 
 #ifdef __INNOTEK_LIBC__
-#define socklen_t               __socklen_t
-#define strcasecmp              stricmp
-#define strncasecmp             strnicmp
-#define NI_NUMERICHOST          1
-#define NI_NUMERICSERV          2
-#define get_last_socket_error() sock_errno()
-#define get_last_error()        errno
-#define readsocket(s,b,n)       recv((s),(b),(n),0)
-#define writesocket(s,b,n)      send((s),(b),(n),0)
-#define closesocket(s)          close(s)
-#define ioctlsocket(a,b,c)      so_ioctl((a),(b),(c))
+#define socklen_t                   __socklen_t
+#define strcasecmp                  stricmp
+#define strncasecmp                 strnicmp
+#define NI_NUMERICHOST              1
+#define NI_NUMERICSERV              2
+#define get_last_socket_error()     sock_errno()
+#define set_last_socket_error(e)    ()
+#define get_last_error()            errno
+#define set_last_error(e)           (errno=(e))
+#define readsocket(s,b,n)           recv((s),(b),(n),0)
+#define writesocket(s,b,n)          send((s),(b),(n),0)
+#define closesocket(s)              close(s)
+#define ioctlsocket(a,b,c)          so_ioctl((a),(b),(c))
 #else
-#define get_last_socket_error() errno
-#define get_last_error()        errno
-#define readsocket(s,b,n)       read((s),(b),(n))
-#define writesocket(s,b,n)      write((s),(b),(n))
-#define closesocket(s)          close(s)
-#define ioctlsocket(a,b,c)      ioctl((a),(b),(c))
+#define get_last_socket_error()     errno
+#define set_last_socket_error(e)    (errno=(e))
+#define get_last_error()            errno
+#define set_last_error(e)           (errno=(e))
+#define readsocket(s,b,n)           read((s),(b),(n))
+#define writesocket(s,b,n)          write((s),(b),(n))
+#define closesocket(s)              close(s)
+#define ioctlsocket(a,b,c)          ioctl((a),(b),(c))
 #endif
 
     /* OpenVMS compatibility */
 #ifdef __vms
-#define libdir "__NA__"
+#define LIBDIR "__NA__"
 #define PIDFILE "SYS$LOGIN:STUNNEL.PID"
 #ifdef __alpha
 #define HOST "alpha-openvms"
@@ -289,6 +320,9 @@ typedef unsigned long u32;
 #include <arpa/inet.h>   /* inet_ntoa */
 #include <sys/time.h>    /* select */
 #include <sys/ioctl.h>   /* ioctl */
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 #include <netinet/tcp.h>
 #include <netdb.h>
 #ifndef INADDR_ANY
@@ -330,6 +364,7 @@ extern char *sys_errlist[];
 #endif /* IP_TRANSPARENT */
 #ifdef HAVE_LINUX_NETFILTER_IPV4_H
 #include <limits.h>
+#include <linux/types.h>
 #include <linux/netfilter_ipv4.h>
 #endif /* HAVE_LINUX_NETFILTER_IPV4_H */
 #endif /* __linux__ */
@@ -338,14 +373,20 @@ extern char *sys_errlist[];
 
 /**************************************** OpenSSL headers */
 
-#ifdef HAVE_OPENSSL
-
 #define OPENSSL_THREAD_DEFINES
 #include <openssl/opensslconf.h>
-#if !defined(OPENSSL_THREADS) && defined(USE_PTHREAD)
+#if defined(USE_PTHREAD) && !(defined(OPENSSL_THREADS) || \
+    (OPENSSL_VERSION_NUMBER<0x0090700fL && defined(THREADS)))
 #error OpenSSL library compiled without thread support
 #endif /* !OPENSSL_THREADS && USE_PTHREAD */
 
+#if defined (USE_WIN32) && defined(OPENSSL_FIPS)
+#define USE_FIPS
+#endif
+
+/* OpenSSL 0.9.6 comp.h needs ZLIB macro to declare COMP_zlib() */
+#define ZLIB
+
 #include <openssl/lhash.h>
 #include <openssl/ssl.h>
 #include <openssl/err.h>
@@ -364,29 +405,23 @@ extern char *sys_errlist[];
 #endif
 #endif /* HAVE_OSSL_ENGINE_H */
 
+#ifdef HAVE_OSSL_OCSP_H
 #include <openssl/ocsp.h>
+#endif /* HAVE_OSSL_OCSP_H */
 
 #ifdef USE_FIPS
 #include <openssl/fips.h>
 #include <openssl/fips_rand.h>
 #endif /* USE_FIPS */
 
-#if OPENSSL_VERSION_NUMBER<0x10000000
+#if OPENSSL_VERSION_NUMBER<0x0090800fL
+#define OPENSSL_NO_ECDH
+#endif /* OpenSSL version < 0.8.0 */
+
+#if OPENSSL_VERSION_NUMBER<0x10000000L
 #define OPENSSL_NO_TLSEXT
 #endif /* OpenSSL version < 1.0.0 */
 
-#else /* HAVE_OPENSSL */
-
-#include <lhash.h>
-#include <ssl.h>
-#include <err.h>
-#include <crypto.h>
-#include <md4.h>
-#include <des.h>
-#define OPENSSL_NO_TLSEXT
-
-#endif /* HAVE_OPENSSL */
-
 /**************************************** other defines */
 
 /* change all non-printable characters to '.' */
@@ -398,14 +433,6 @@ extern char *sys_errlist[];
     do {unsigned char *p; for(p=(s); *p; p++) \
         if(!isalnum((int)*p)) *p='.';} while(0)
 
-/* some definitions for IPv6 support */
-#if defined(USE_IPv6)
-#define addr_len(x) ((x).sa.sa_family==AF_INET ? \
-    sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6))
-#else
-#define addr_len(x) (sizeof(struct sockaddr_in))
-#endif
-
 /* always use IPv4 defaults! */
 #define DEFAULT_LOOPBACK "127.0.0.1"
 #define DEFAULT_ANY "0.0.0.0"
