@@ -40,25 +40,27 @@
 
 /* #define DEBUG_UCONTEXT */
 
-static int signal_pipe[2]={-1, -1};
-#ifdef __INNOTEK_LIBC__
-struct sockaddr_un {
-    u_char  sun_len;             /* sockaddr len including null */
-    u_char  sun_family;          /* AF_OS2 or AF_UNIX */
-    char    sun_path[108];       /* path name */
-};
-#endif
-static void signal_pipe_empty(void);
-#ifdef USE_FORK
-static void client_status(void); /* dead children detected */
-#endif
-
 /**************************************** s_poll functions */
 
 #ifdef USE_POLL
 
+s_poll_set *s_poll_alloc() {
+    /* it needs to be filled with zeros */
+    return str_alloc(sizeof(s_poll_set));
+}
+
+void s_poll_free(s_poll_set *fds) {
+    if(fds) {
+        if(fds->ufds)
+            str_free(fds->ufds);
+        str_free(fds);
+    }
+}
+
 void s_poll_init(s_poll_set *fds) {
     fds->nfds=0;
+    fds->allocated=4; /* prealloc 4 file desciptors */
+    fds->ufds=str_realloc(fds->ufds, fds->allocated*sizeof(struct pollfd));
 }
 
 void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
@@ -66,12 +68,11 @@ void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
 
     for(i=0; i<fds->nfds && fds->ufds[i].fd!=fd; i++)
         ;
-    if(i>=MAX_FD) {
-        s_log(LOG_ERR,
-            "s_poll_add failed for FD=%d: too many file descriptors", fd);
-        return;
-    }
     if(i==fds->nfds) {
+        if(i==fds->allocated) {
+            fds->allocated=i+1;
+            fds->ufds=str_realloc(fds->ufds, fds->allocated*sizeof(struct pollfd));
+        }
         fds->ufds[i].fd=fd;
         fds->ufds[i].events=0;
         fds->nfds++;
@@ -114,12 +115,11 @@ int s_poll_error(s_poll_set *fds, int fd) {
 
 /* move ready contexts from waiting queue to ready queue */
 static void scan_waiting_queue(void) {
-    int retval, retry;
+    int retval;
     CONTEXT *context, *prev;
     int min_timeout;
     unsigned int nfds, i;
     time_t now;
-    short *signal_revents;
     static unsigned int max_nfds=0;
     static struct pollfd *ufds=NULL;
 
@@ -136,20 +136,13 @@ static void scan_waiting_queue(void) {
     /* setup ufds structure */
     if(nfds>max_nfds) { /* need to allocate more memory */
         ufds=str_realloc(ufds, nfds*sizeof(struct pollfd));
-        if(!ufds) {
-            s_log(LOG_CRIT, "Memory allocation failed");
-            die(1);
-        }
         max_nfds=nfds;
     }
     nfds=0;
-    signal_revents=NULL;
     for(context=waiting_head; context; context=context->next)
         for(i=0; i<context->fds->nfds; i++) {
             ufds[nfds].fd=context->fds->ufds[i].fd;
             ufds[nfds].events=context->fds->ufds[i].events;
-            if(ufds[nfds].fd==signal_pipe[0])
-                signal_revents=&ufds[nfds].revents;
             nfds++;
         }
 
@@ -158,13 +151,8 @@ static void scan_waiting_queue(void) {
         min_timeout, nfds);
 #endif
     do { /* skip "Interrupted system call" errors */
-        retry=0;
         retval=poll(ufds, nfds, min_timeout<0 ? -1 : 1000*min_timeout);
-        if(retval>0 && signal_revents && (*signal_revents & POLLIN)) {
-            signal_pipe_empty(); /* no timeout -> main loop */
-            retry=1;
-        }
-    } while(retry || (retval<0 && get_last_socket_error()==EINTR));
+    } while(retval<0 && get_last_socket_error()==S_EINTR);
     time(&now);
     /* process the returned data */
     nfds=0;
@@ -229,9 +217,11 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
 
     /* it's illegal to deallocate the stack of the current context */
     if(to_free) { /* a delayed deallocation is scheduled */
+#ifdef DEBUG_UCONTEXT
         s_log(LOG_DEBUG, "Releasing context %ld", to_free->id);
-        free(to_free->stack);
-        free(to_free);
+#endif
+        str_free(to_free->stack);
+        str_free(to_free);
         to_free=NULL;
     }
 
@@ -257,15 +247,21 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     /* switch threads */
     if(fds) { /* swap the current context */
         if(context->id!=ready_head->id) {
+#ifdef DEBUG_UCONTEXT
             s_log(LOG_DEBUG, "Context swap: %ld -> %ld",
                 context->id, ready_head->id);
+#endif
             swapcontext(&context->context, &ready_head->context);
+#ifdef DEBUG_UCONTEXT
             s_log(LOG_DEBUG, "Current context: %ld", ready_head->id);
+#endif
         }
         return ready_head->ready;
     } else { /* drop the current context */
+#ifdef DEBUG_UCONTEXT
         s_log(LOG_DEBUG, "Context set: %ld (dropped) -> %ld",
             context->id, ready_head->id);
+#endif
         setcontext(&ready_head->context);
         ioerror("setcontext"); /* should not ever happen */
         return 0;
@@ -275,16 +271,11 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
 #else /* USE_UCONTEXT */
 
 int s_poll_wait(s_poll_set *fds, int sec, int msec) {
-    int retval, retry;
+    int retval;
 
     do { /* skip "Interrupted system call" errors */
-        retry=0;
         retval=poll(fds->ufds, fds->nfds, sec<0 ? -1 : 1000*sec+msec);
-        if(sec<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
-            signal_pipe_empty(); /* no timeout -> main loop */
-            retry=1;
-        }
-    } while(retry || (retval<0 && get_last_socket_error()==EINTR));
+    } while(retval<0 && get_last_socket_error()==S_EINTR);
     return retval;
 }
 
@@ -292,6 +283,16 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
 
 #else /* select */
 
+s_poll_set *s_poll_alloc() {
+    /* it needs to be filled with zeros */
+    return str_alloc(sizeof(s_poll_set));
+}
+
+void s_poll_free(s_poll_set *fds) {
+    if(fds)
+        str_free(fds);
+}
+
 void s_poll_init(s_poll_set *fds) {
     FD_ZERO(&fds->irfds);
     FD_ZERO(&fds->iwfds);
@@ -322,11 +323,10 @@ int s_poll_error(s_poll_set *fds, int fd) {
 }
 
 int s_poll_wait(s_poll_set *fds, int sec, int msec) {
-    int retval, retry;
+    int retval;
     struct timeval tv, *tv_ptr;
 
     do { /* skip "Interrupted system call" errors */
-        retry=0;
         memcpy(&fds->orfds, &fds->irfds, sizeof(fd_set));
         memcpy(&fds->owfds, &fds->iwfds, sizeof(fd_set));
         if(sec<0) { /* infinite timeout */
@@ -337,186 +337,12 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
             tv_ptr=&tv;
         }
         retval=select(fds->max+1, &fds->orfds, &fds->owfds, NULL, tv_ptr);
-        if(sec<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
-            signal_pipe_empty(); /* no timeout -> main loop */
-            retry=1;
-        }
-    } while(retry || (retval<0 && get_last_socket_error()==EINTR));
+    } while(retval<0 && get_last_socket_error()==S_EINTR);
     return retval;
 }
 
 #endif /* USE_POLL */
 
-/**************************************** signal pipe handling */
-
-int signal_pipe_init(void) {
-#ifdef USE_WIN32
-    if(make_sockets(signal_pipe))
-        die(1);
-#else
-#if defined(__INNOTEK_LIBC__)
-    /* Innotek port of GCC can not use select on a pipe
-     * use local socket instead */
-    struct sockaddr_un un;
-    fd_set set_pipe;
-    int pipe_in;
-
-    FD_ZERO(&set_pipe);
-    signal_pipe[0]=s_socket(PF_OS2, SOCK_STREAM, 0, 0, "socket#1");
-    signal_pipe[1]=s_socket(PF_OS2, SOCK_STREAM, 0, 0, "socket#2");
-
-    /* connect the two endpoints */
-    memset(&un, 0, sizeof un);
-    un.sun_len=sizeof un;
-    un.sun_family=AF_OS2;
-    sprintf(un.sun_path, "\\socket\\stunnel-%u", getpid());
-    /* make the first endpoint listen */
-    bind(signal_pipe[0], (struct sockaddr *)&un, sizeof un);
-    listen(signal_pipe[0], 1);
-    connect(signal_pipe[1], (struct sockaddr *)&un, sizeof un);
-    FD_SET(signal_pipe[0], &set_pipe);
-    if(select(signal_pipe[0]+1, &set_pipe, NULL, NULL, NULL)>0) {
-        pipe_in=signal_pipe[0];
-        signal_pipe[0]=s_accept(signal_pipe[0], NULL, 0, 0, "accept");
-        closesocket(pipe_in);
-    } else {
-        sockerror("select");
-        die(1);
-    }
-#else /* __INNOTEK_LIBC__ */
-    if(s_pipe(signal_pipe, 1, "signal_pipe"))
-        die(1);
-#endif /* __INNOTEK_LIBC__ */
-
-    signal(SIGCHLD, signal_handler); /* a child has died */
-    signal(SIGHUP, signal_handler); /* configuration reload */
-    signal(SIGUSR1, signal_handler); /* log reopen */
-    signal(SIGPIPE, SIG_IGN); /* ignore "broken pipe" */
-    if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
-        signal(SIGTERM, signal_handler); /* fatal */
-    if(signal(SIGQUIT, SIG_IGN)!=SIG_IGN)
-        signal(SIGQUIT, signal_handler); /* fatal */
-    if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
-        signal(SIGINT, signal_handler); /* fatal */
-    /* signal(SIGSEGV, signal_handler); */
-#endif /* USE_WIN32 */
-    return signal_pipe[0];
-}
-
-void signal_post(int sig) {
-    writesocket(signal_pipe[1], (char *)&sig, sizeof sig);
-}
-
-static void signal_pipe_empty(void) {
-    int sig, err;
-
-    s_log(LOG_DEBUG, "Dispatching signals from the signal pipe");
-    while(readsocket(signal_pipe[0], (char *)&sig, sizeof sig)==sizeof sig) {
-        switch(sig) {
-#ifndef USE_WIN32
-        case SIGCHLD:
-            s_log(LOG_DEBUG, "Processing SIGCHLD");
-#ifdef USE_FORK
-            client_status(); /* report status of client process */
-#else /* USE_UCONTEXT || USE_PTHREAD */
-            child_status();  /* report status of libwrap or 'exec' process */
-#endif /* defined USE_FORK */
-            break;
-#endif /* !defind USE_WIN32 */
-        case SIGNAL_RELOAD_CONFIG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
-            unbind_ports();
-            log_close();
-            err=parse_conf(NULL, CONF_RELOAD);
-            if(err) {
-                log_flush(LOG_MODE_ERROR);
-            } else {
-                log_open();
-                bind_ports();
-            }
-#ifdef USE_WIN32
-            win_newconfig(err);
-#endif
-            break;
-        case SIGNAL_REOPEN_LOG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
-            log_close();
-            log_open();
-            s_log(LOG_NOTICE, "Log file reopened");
-            break;
-        default:
-            s_log(sig==SIGNAL_TERMINATE ? LOG_NOTICE : LOG_ERR,
-                "Received signal %d; terminating", sig);
-            die(3);
-        }
-    }
-    s_log(LOG_DEBUG, "Signal pipe is empty");
-}
-
-#ifdef USE_FORK
-static void client_status(void) { /* dead children detected */
-    int pid, status;
-
-#ifdef HAVE_WAIT_FOR_PID
-    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-        --num_clients; /* one client less */
-#else
-    if((pid=wait(&status))>0) {
-        --num_clients; /* one client less */
-#endif
-#ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            s_log(LOG_DEBUG, "Process %d terminated on signal %d (%d left)",
-                pid, WTERMSIG(status), num_clients);
-        } else {
-            s_log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
-                pid, WEXITSTATUS(status), num_clients);
-        }
-    }
-#else
-        s_log(LOG_DEBUG, "Process %d finished with code %d (%d left)",
-            pid, status, num_clients);
-    }
-#endif
-}
-#endif /* defined USE_FORK */
-
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-
-void child_status(void) { /* dead libwrap or 'exec' process detected */
-    int pid, status;
-
-#ifdef HAVE_WAIT_FOR_PID
-    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-#else
-    if((pid=wait(&status))>0) {
-#endif
-#ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            s_log(LOG_INFO, "Child process %d terminated on signal %d",
-                pid, WTERMSIG(status));
-        } else {
-            s_log(LOG_INFO, "Child process %d finished with code %d",
-                pid, WEXITSTATUS(status));
-        }
-#else
-        s_log(LOG_INFO, "Child process %d finished with status %d",
-            pid, status);
-#endif
-    }
-}
-
-void signal_handler(int sig) {
-    int saved_errno;
-
-    saved_errno=errno;
-    signal_post(sig);
-    signal(sig, signal_handler);
-    errno=saved_errno;
-}
-
-#endif /* !defined(USE_WIN32) && !defined(USE_OS2) */
-
 /**************************************** fd management */
 
 int set_socket_options(int s, int type) {
@@ -543,9 +369,10 @@ int set_socket_options(int s, int type) {
         }
         if(setsockopt(s, ptr->opt_level, ptr->opt_name,
                 (void *)ptr->opt_val[type], opt_size)) {
-            if(get_last_socket_error()==EOPNOTSUPP) {
-                s_log(LOG_NOTICE,
-                    "Failed to set option %s set on %s socket (not a socket?)",
+            if(get_last_socket_error()==S_EOPNOTSUPP) {
+                /* most likely stdin/stdout or AF_UNIX socket */
+                s_log(LOG_DEBUG,
+                    "Option %s not supported on %s socket",
                     ptr->opt_str, type_str[type]);
             } else {
                 sockerror(ptr->opt_str);
@@ -572,53 +399,60 @@ int get_socket_error(const int fd) {
 
 int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
     int error;
-    char dst[IPLEN];
+    char *dst;
 
-    s_ntop(dst, addr);
+    dst=s_ntop(addr, addrlen);
     s_log(LOG_INFO, "connect_blocking: connecting %s", dst);
 
     if(!connect(c->fd, &addr->sa, addrlen)) {
         s_log(LOG_NOTICE, "connect_blocking: connected %s", dst);
+        str_free(dst);
         return 0; /* no error -> success (on some OSes over the loopback) */
     }
     error=get_last_socket_error();
-    if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
+    if(error!=S_EINPROGRESS && error!=S_EWOULDBLOCK) {
         s_log(LOG_ERR, "connect_blocking: connect %s: %s (%d)",
             dst, s_strerror(error), error);
+        str_free(dst);
         return -1;
     }
 
     s_log(LOG_DEBUG, "connect_blocking: s_poll_wait %s: waiting %d seconds",
         dst, c->opt->timeout_connect);
-    s_poll_init(&c->fds);
-    s_poll_add(&c->fds, c->fd, 1, 1);
-    switch(s_poll_wait(&c->fds, c->opt->timeout_connect, 0)) {
+    s_poll_init(c->fds);
+    s_poll_add(c->fds, c->fd, 1, 1);
+    switch(s_poll_wait(c->fds, c->opt->timeout_connect, 0)) {
     case -1:
         error=get_last_socket_error();
         s_log(LOG_ERR, "connect_blocking: s_poll_wait %s: %s (%d)",
             dst, s_strerror(error), error);
+        str_free(dst);
         return -1;
     case 0:
         s_log(LOG_ERR, "connect_blocking: s_poll_wait %s:"
             " TIMEOUTconnect exceeded", dst);
+        str_free(dst);
         return -1;
     default:
-        if(s_poll_canread(&c->fds, c->fd) || s_poll_error(&c->fds, c->fd)) {
+        if(s_poll_canread(c->fds, c->fd) || s_poll_error(c->fds, c->fd)) {
             /* newly connected socket should not be ready for read */
             /* get the resulting error code, now */
             error=get_socket_error(c->fd);
             if(error) { /* really an error? */
                 s_log(LOG_ERR, "connect_blocking: getsockopt %s: %s (%d)",
                     dst, s_strerror(error), error);
+                str_free(dst);
                 return -1;
             }
         }
-        if(s_poll_canwrite(&c->fds, c->fd)) {
+        if(s_poll_canwrite(c->fds, c->fd)) {
             s_log(LOG_NOTICE, "connect_blocking: connected %s", dst);
+            str_free(dst);
             return 0; /* success */
         }
         s_log(LOG_ERR, "connect_blocking: s_poll_wait %s: internal error",
             dst);
+        str_free(dst);
         return -1;
     }
     return -1; /* should not be possible */
@@ -626,13 +460,12 @@ int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
 
 void write_blocking(CLI *c, int fd, void *ptr, int len) {
         /* simulate a blocking write */
-    s_poll_set fds;
     int num;
 
     while(len>0) {
-        s_poll_init(&fds);
-        s_poll_add(&fds, fd, 0, 1); /* write */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
+        s_poll_init(c->fds);
+        s_poll_add(c->fds, fd, 0, 1); /* write */
+        switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
         case -1:
             sockerror("write_blocking: s_poll_wait");
             longjmp(c->err, 1); /* error */
@@ -659,13 +492,12 @@ void write_blocking(CLI *c, int fd, void *ptr, int len) {
 
 void read_blocking(CLI *c, int fd, void *ptr, int len) {
         /* simulate a blocking read */
-    s_poll_set fds;
     int num;
 
     while(len>0) {
-        s_poll_init(&fds);
-        s_poll_add(&fds, fd, 1, 0); /* read */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
+        s_poll_init(c->fds);
+        s_poll_add(c->fds, fd, 1, 0); /* read */
+        switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
         case -1:
             sockerror("read_blocking: s_poll_wait");
             longjmp(c->err, 1); /* error */
@@ -693,7 +525,7 @@ void read_blocking(CLI *c, int fd, void *ptr, int len) {
     }
 }
 
-void fdputline(CLI *c, int fd, const char *line) {
+void fd_putline(CLI *c, int fd, const char *line) {
     char *tmpline;
     const char crlf[]="\r\n";
     int len;
@@ -707,43 +539,38 @@ void fdputline(CLI *c, int fd, const char *line) {
     str_free(tmpline);
 }
 
-char *fdgetline(CLI *c, int fd) {
+char *fd_getline(CLI *c, int fd) {
     char *line=NULL, *tmpline;
-    s_poll_set fds;
     int ptr=0;
 
     for(;;) {
-        s_poll_init(&fds);
-        s_poll_add(&fds, fd, 1, 0); /* read */
-        switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
+        s_poll_init(c->fds);
+        s_poll_add(c->fds, fd, 1, 0); /* read */
+        switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
         case -1:
-            sockerror("fdgetline: s_poll_wait");
+            sockerror("fd_getline: s_poll_wait");
             str_free(line);
             longjmp(c->err, 1); /* error */
         case 0:
-            s_log(LOG_INFO, "fdgetline: s_poll_wait:"
+            s_log(LOG_INFO, "fd_getline: s_poll_wait:"
                 " TIMEOUTbusy exceeded: sending reset");
             str_free(line);
             longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
-            s_log(LOG_ERR, "fdgetline: s_poll_wait: unknown result");
+            s_log(LOG_ERR, "fd_getline: s_poll_wait: Unknown result");
             str_free(line);
             longjmp(c->err, 1); /* error */
         }
         line=str_realloc(line, ptr+1);
-        if(!line) {
-            s_log(LOG_CRIT, "Memory allocation failed");
-            longjmp(c->err, 1); /* error */
-        }
         switch(readsocket(fd, line+ptr, 1)) {
         case -1: /* error */
-            sockerror("readsocket (fdgetline)");
+            sockerror("fd_getline: readsocket");
             str_free(line);
             longjmp(c->err, 1);
         case 0: /* EOF */
-            s_log(LOG_ERR, "Unexpected socket close (fdgetline)");
+            s_log(LOG_ERR, "fd_getline: Unexpected socket close");
             str_free(line);
             longjmp(c->err, 1);
         }
@@ -751,9 +578,13 @@ char *fdgetline(CLI *c, int fd) {
             continue;
         if(line[ptr]=='\n')
             break;
-        if(!line[ptr])
+        if(line[ptr]=='\0')
             break;
-        ++ptr;
+        if(++ptr>65536) { /* >64KB --> DoS protection */
+            s_log(LOG_ERR, "fd_getline: Line too long");
+            str_free(line);
+            longjmp(c->err, 1);
+        }
     }
     line[ptr]='\0';
     tmpline=str_dup(line);
@@ -763,7 +594,7 @@ char *fdgetline(CLI *c, int fd) {
     return line;
 }
 
-int fdprintf(CLI *c, int fd, const char *format, ...) {
+void fd_printf(CLI *c, int fd, const char *format, ...) {
     va_list ap;
     char *line;
 
@@ -771,12 +602,11 @@ int fdprintf(CLI *c, int fd, const char *format, ...) {
     line=str_vprintf(format, ap);
     va_end(ap);
     if(!line) {
-        s_log(LOG_ERR, "fdprintf: str_vprintf failed");
+        s_log(LOG_ERR, "fd_printf: str_vprintf failed");
         longjmp(c->err, 1);
     }
-    fdputline(c, fd, line);
+    fd_putline(c, fd, line);
     str_free(line);
-    return strlen(line)+2;
 }
 
 #define INET_SOCKET_PAIR
