@@ -53,23 +53,38 @@
 static int section_init(int, SERVICE_OPTIONS *, int);
 
 static int parse_debug_level(char *);
+
 static int parse_ssl_option(char *);
+
 static int print_socket_options(void);
 static char *print_option(int, OPT_UNION *);
 static int parse_socket_option(char *);
+
+#ifdef HAVE_OSSL_OCSP_H
 static char *parse_ocsp_url(SERVICE_OPTIONS *, char *);
 static unsigned long parse_ocsp_flag(char *);
+#endif /* HAVE_OSSL_OCSP_H */
+
+#ifdef HAVE_OSSL_ENGINE_H
+static char *open_engine(const char *);
+static char *ctrl_engine(const char *, const char *);
+static char *init_engine(void);
+static void close_engine(void);
+static ENGINE *get_engine(int);
+#endif
 
 static void print_syntax(void);
 static void config_error(int, const char *, const char *);
 static void section_error(int, const char *, const char *);
-static char *str_dup_err(char *);
 #ifndef USE_WIN32
 static char **argalloc(char *);
 #endif
 
-GLOBAL_OPTIONS global_options, new_global_options;
-SERVICE_OPTIONS service_options, new_service_options;
+GLOBAL_OPTIONS global_options;
+SERVICE_OPTIONS service_options;
+
+static GLOBAL_OPTIONS new_global_options;
+static SERVICE_OPTIONS new_service_options;
 
 typedef enum {
     CMD_INIT, /* initialize */
@@ -90,8 +105,10 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     struct passwd *pw;
 #endif
 
-    if(cmd==CMD_DEFAULT || cmd==CMD_HELP)
-        s_log(LOG_NOTICE, "Global option defaults");
+    if(cmd==CMD_DEFAULT || cmd==CMD_HELP) {
+        s_log(LOG_NOTICE, " ");
+        s_log(LOG_NOTICE, "Global options:");
+    }
 
     /* chroot */
 #ifdef HAVE_CHROOT
@@ -102,7 +119,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "chroot"))
             break;
-        new_global_options.chroot_dir=str_dup_err(arg);
+        new_global_options.chroot_dir=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -146,7 +163,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "debug"))
             break;
-        if(!parse_debug_level(arg))
+        if(parse_debug_level(arg))
             return "Illegal debug argument";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -173,7 +190,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "EGD"))
             break;
-        new_global_options.egd_sock=str_dup_err(arg);
+        new_global_options.egd_sock=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
 #ifdef EGD_SOCKET
@@ -280,7 +297,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "output"))
             break;
-        new_global_options.output_file=str_dup_err(arg);
+        new_global_options.output_file=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -299,7 +316,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
         if(strcasecmp(opt, "pid"))
             break;
         if(arg[0]) /* is argument not empty? */
-            new_global_options.pidfile=str_dup_err(arg);
+            new_global_options.pidfile=str_dup(arg);
         else
             new_global_options.pidfile=NULL; /* empty -> do not create a pid file */
         return NULL; /* OK */
@@ -340,7 +357,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "RNDfile"))
             break;
-        new_global_options.rand_file=str_dup_err(arg);
+        new_global_options.rand_file=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
 #ifdef RANDOM_FILE
@@ -376,15 +393,16 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
         break;
     }
 
+#ifndef USE_WIN32
     /* service */
     switch(cmd) {
     case CMD_INIT:
-        new_service_options.servname=str_dup_err("stunnel");
+        new_service_options.servname=str_dup("stunnel");
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "service"))
             break;
-        new_service_options.servname=str_dup_err(arg);
+        new_service_options.servname=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -392,6 +410,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
         s_log(LOG_NOTICE, "%-15s = service name", "service");
         break;
     }
+#endif
 
 #ifndef USE_WIN32
     /* setgid */
@@ -452,7 +471,7 @@ static char *parse_global_option(CMD cmd, char *opt, char *arg) {
     case CMD_EXEC:
         if(strcasecmp(opt, "socket"))
             break;
-        if(!parse_socket_option(arg))
+        if(parse_socket_option(arg))
             return "Illegal socket option";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -530,22 +549,22 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
 
     if(cmd==CMD_DEFAULT || cmd==CMD_HELP) {
         s_log(LOG_NOTICE, " ");
-        s_log(LOG_NOTICE, "Service-level option defaults");
+        s_log(LOG_NOTICE, "Service-level options:");
     }
 
     /* accept */
     switch(cmd) {
     case CMD_INIT:
         section->option.accept=0;
-        memset(&section->local_addr, 0, sizeof(SOCKADDR_LIST));
-        section->local_addr.addr[0].in.sin_family=AF_INET;
+        memset(&section->local_addr, 0, sizeof(SOCKADDR_UNION));
+        section->local_addr.in.sin_family=AF_INET;
         section->fd=-1;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "accept"))
             break;
         section->option.accept=1;
-        if(!name2addrlist(&section->local_addr, arg, DEFAULT_ANY))
+        if(!name2addr(&section->local_addr, arg, DEFAULT_ANY))
             return "Failed to resolve accepting address";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -568,7 +587,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "CApath"))
             break;
         if(arg[0]) /* not empty */
-            section->ca_dir=str_dup_err(arg);
+            section->ca_dir=str_dup(arg);
         else
             section->ca_dir=NULL;
         return NULL; /* OK */
@@ -596,7 +615,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "CAfile"))
             break;
         if(arg[0]) /* not empty */
-            section->ca_file=str_dup_err(arg);
+            section->ca_file=str_dup(arg);
         else
             section->ca_file=NULL;
         return NULL; /* OK */
@@ -620,7 +639,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "cert"))
             break;
-        section->cert=str_dup_err(arg);
+        section->cert=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break; /* no default certificate */
@@ -642,7 +661,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "ciphers"))
             break;
-        section->cipher_list=str_dup_err(arg);
+        section->cipher_list=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         s_log(LOG_NOTICE, "%-15s = %s", "ciphers", STUNNEL_CIPHER_LIST);
@@ -679,27 +698,19 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     switch(cmd) {
     case CMD_INIT:
         section->option.remote=0;
-        section->remote_address=NULL;
-        section->remote_addr.num=0;
-        section->host_name=NULL;
+        section->connect_name=NULL;
+        section->connect_addr.num=0;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "connect"))
             break;
         section->option.remote=1;
-        section->remote_address=str_dup_err(arg);
+        section->connect_name=str_dup(arg);
         if(!section->option.delayed_lookup &&
-                !name2addrlist(&section->remote_addr, arg, DEFAULT_LOOPBACK)) {
+                !name2addrlist(&section->connect_addr, arg, DEFAULT_LOOPBACK)) {
             s_log(LOG_INFO, "Cannot resolve '%s' - delaying DNS lookup", arg);
             section->option.delayed_lookup=1;
         }
-        tmpstr=strrchr(arg, ':');
-        if(tmpstr) {
-            *tmpstr='\0';
-            section->host_name=str_dup_err(arg);
-        } else {
-            section->host_name=str_dup_err("localhost");
-        }
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -718,7 +729,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "CRLpath"))
             break;
         if(arg[0]) /* not empty */
-            section->crl_dir=str_dup_err(arg);
+            section->crl_dir=str_dup(arg);
         else
             section->crl_dir=NULL;
         return NULL; /* OK */
@@ -738,7 +749,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "CRLfile"))
             break;
         if(arg[0]) /* not empty */
-            section->crl_file=str_dup_err(arg);
+            section->crl_file=str_dup(arg);
         else
             section->crl_file=NULL;
         return NULL; /* OK */
@@ -749,6 +760,8 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         break;
     }
 
+#ifndef OPENSSL_NO_ECDH
+
     /* curve */
 #define DEFAULT_CURVE NID_X9_62_prime256v1
     switch(cmd) {
@@ -770,6 +783,8 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         break;
     }
 
+#endif /* !OPENSSL_NO_ECDH */
+
     /* delay */
     switch(cmd) {
     case CMD_INIT:
@@ -827,9 +842,9 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "exec"))
             break;
         section->option.program=1;
-        section->execname=str_dup_err(arg);
+        section->execname=str_dup(arg);
 #ifdef USE_WIN32
-        section->execargs=str_dup_err(arg);
+        section->execargs=str_dup(arg);
 #else
         if(!section->execargs) {
             section->execargs=str_alloc(2*sizeof(char *));
@@ -855,7 +870,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         if(strcasecmp(opt, "execargs"))
             break;
 #ifdef USE_WIN32
-        section->execargs=str_dup_err(arg);
+        section->execargs=str_dup(arg);
 #else
         section->execargs=argalloc(arg);
 #endif
@@ -899,7 +914,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "ident"))
             break;
-        section->username=str_dup_err(arg);
+        section->username=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -916,7 +931,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "key"))
             break;
-        section->key=str_dup_err(arg);
+        section->key=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -952,13 +967,15 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     /* local */
     switch(cmd) {
     case CMD_INIT:
-        memset(&section->source_addr, 0, sizeof(SOCKADDR_LIST));
-        section->source_addr.addr[0].in.sin_family=AF_INET;
+        section->option.local=0;
+        memset(&section->source_addr, 0, sizeof(SOCKADDR_UNION));
+        section->source_addr.in.sin_family=AF_INET;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "local"))
             break;
-        if(!hostport2addrlist(&section->source_addr, arg, "0"))
+        section->option.local=1;
+        if(!hostport2addr(&section->source_addr, arg, "0"))
             return "Failed to resolve local address";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -969,12 +986,14 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         break;
     }
 
+#ifdef HAVE_OSSL_OCSP_H
+
     /* OCSP */
     switch(cmd) {
     case CMD_INIT:
         section->option.ocsp=0;
-        memset(&section->ocsp_addr, 0, sizeof(SOCKADDR_LIST));
-        section->ocsp_addr.addr[0].in.sin_family=AF_INET;
+        memset(&section->ocsp_addr, 0, sizeof(SOCKADDR_UNION));
+        section->ocsp_addr.in.sin_family=AF_INET;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "ocsp"))
@@ -1008,6 +1027,8 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         break;
     }
 
+#endif /* HAVE_OSSL_OCSP_H */
+
     /* options */
     switch(cmd) {
     case CMD_INIT:
@@ -1032,19 +1053,21 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     /* protocol */
     switch(cmd) {
     case CMD_INIT:
-        section->protocol=NULL;
+        section->protocol=-1;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "protocol"))
             break;
-        section->protocol=str_dup_err(arg);
+        section->protocol=find_protocol_id(arg);
+        if(section->protocol<0)
+            return "Unknown protocol";
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
     case CMD_HELP:
         s_log(LOG_NOTICE, "%-15s = protocol to negotiate before SSL initialization",
             "protocol");
-        s_log(LOG_NOTICE, "%18scurrently supported: cifs, connect, nntp, pop3, smtp", "");
+        s_log(LOG_NOTICE, "%18scurrently supported: cifs, connect, imap, nntp, pgsql, pop3, proxy, smtp", "");
         break;
     }
 
@@ -1056,7 +1079,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "protocolAuthentication"))
             break;
-        section->protocol_authentication=str_dup_err(arg);
+        section->protocol_authentication=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -1074,7 +1097,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "protocolHost"))
             break;
-        section->protocol_host=str_dup_err(arg);
+        section->protocol_host=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -1092,7 +1115,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "protocolPassword"))
             break;
-        section->protocol_password=str_dup_err(arg);
+        section->protocol_password=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -1110,7 +1133,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_EXEC:
         if(strcasecmp(opt, "protocolUsername"))
             break;
-        section->protocol_username=str_dup_err(arg);
+        section->protocol_username=str_dup(arg);
         return NULL; /* OK */
     case CMD_DEFAULT:
         break;
@@ -1192,8 +1215,8 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     switch(cmd) {
     case CMD_INIT:
         section->option.sessiond=0;
-        memset(&section->sessiond_addr, 0, sizeof(SOCKADDR_LIST));
-        section->sessiond_addr.addr[0].in.sin_family=AF_INET;
+        memset(&section->sessiond_addr, 0, sizeof(SOCKADDR_UNION));
+        section->sessiond_addr.in.sin_family=AF_INET;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "sessiond"))
@@ -1204,7 +1227,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         /* this prevents session callbacks from beeing executed */
         section->ssl_options|=SSL_OP_NO_TICKET;
 #endif
-        if(!name2addrlist(&section->sessiond_addr, arg, DEFAULT_LOOPBACK))
+        if(!name2addr(&section->sessiond_addr, arg, DEFAULT_LOOPBACK))
             return "Failed to resolve sessiond server address";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -1246,11 +1269,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
                 str_alloc(sizeof(SERVERNAME_LIST));
             tmpsrv->ssl_options|=SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
         }
-        if(!tmpsrv->servername_list_tail) {
-            s_log(LOG_ERR, "Fatal memory allocation error");
-            die(2);
-        }
-        tmpsrv->servername_list_tail->servername=str_dup_err(tmpstr);
+        tmpsrv->servername_list_tail->servername=str_dup(tmpstr);
         tmpsrv->servername_list_tail->opt=section;
         tmpsrv->servername_list_tail->next=NULL;
         section->option.sni=1;
@@ -1494,7 +1513,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         section->verify_level=strtol(arg, &tmpstr, 10);
         if(tmpstr==arg || *tmpstr) /* not a number */
             return "Bad verify level";
-        if(section->verify_level<0 || section->verify_level>3)
+        if(section->verify_level<0 || section->verify_level>4)
             return "Bad verify level";
         return NULL; /* OK */
     case CMD_DEFAULT:
@@ -1511,6 +1530,8 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
             "%18slevel 2 - always require a valid peer certificate", "");
         s_log(LOG_NOTICE,
             "%18slevel 3 - verify peer with locally installed certificate", "");
+        s_log(LOG_NOTICE,
+            "%18slevel 4 - ignore CA chain and only verify peer certificate", "");
         break;
     }
 
@@ -1521,7 +1542,7 @@ static char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
 
 /**************************************** parse commandline parameters */
 
-void parse_commandline(char *name, char *parameter) {
+int parse_commandline(char *name, char *parameter) {
     if(!name)
 #ifdef CONFDIR
         name=CONFDIR CONFSEPARATOR "stunnel.conf";
@@ -1532,20 +1553,21 @@ void parse_commandline(char *name, char *parameter) {
     if(!strcasecmp(name, "-help")) {
         parse_global_option(CMD_HELP, NULL, NULL);
         parse_service_option(CMD_HELP, NULL, NULL, NULL);
-        die(1);
+        log_flush(LOG_MODE_INFO);
+        return 1;
     }
 
     if(!strcasecmp(name, "-version")) {
-        stunnel_info(LOG_NOTICE);
-        s_log(LOG_NOTICE, " ");
         parse_global_option(CMD_DEFAULT, NULL, NULL);
         parse_service_option(CMD_DEFAULT, NULL, NULL, NULL);
-        die(1);
+        log_flush(LOG_MODE_INFO);
+        return 1;
     }
 
     if(!strcasecmp(name, "-sockets")) {
         print_socket_options();
-        die(1);
+        log_flush(LOG_MODE_INFO);
+        return 1;
     }
 
 #ifndef USE_WIN32
@@ -1553,16 +1575,18 @@ void parse_commandline(char *name, char *parameter) {
         if(!parameter) {
             s_log(LOG_ERR, "No file descriptor specified");
             print_syntax();
-            die(1);
+            return 1;
         }
         if(parse_conf(parameter, CONF_FD))
-            die(1);
+            return 1;
     } else
 #else
     (void)parameter; /* skip warning about unused parameter */
 #endif
         if(parse_conf(name, CONF_FILE))
-            die(1);
+            return 1;
+    apply_conf();
+    return 0;
 }
 
 /**************************************** parse configuration file */
@@ -1574,16 +1598,13 @@ int parse_conf(char *name, CONF_TYPE type) {
     int line_number, i;
     SERVICE_OPTIONS *section, *new_section;
     static char *filename=NULL; /* a copy of config file name for reloading */
-#ifdef MAX_FD
-    int sections=0;
-#endif
 #ifndef USE_WIN32
     int fd;
     char *tmpstr;
 #endif
 
     if(name) /* not reload */
-        filename=str_dup_err(name);
+        filename=str_dup(name);
 
     s_log(LOG_NOTICE, "Reading configuration from %s %s",
         type==CONF_FD ? "descriptor" : "file", filename);
@@ -1593,7 +1614,7 @@ int parse_conf(char *name, CONF_TYPE type) {
         if(tmpstr==filename || *tmpstr) { /* not a number */
             s_log(LOG_ERR, "Invalid file descriptor number");
             print_syntax();
-            die(1);
+            return 1;
         }
         df=file_fdopen(fd);
     } else
@@ -1601,12 +1622,9 @@ int parse_conf(char *name, CONF_TYPE type) {
         df=file_open(filename, 0);
     if(!df) {
         s_log(LOG_ERR, "Cannot read configuration");
-        if(type==CONF_RELOAD) {
-            return 1;
-        } else {
+        if(type!=CONF_RELOAD)
             print_syntax();
-            die(1);
-        }
+        return 1;
     }
 
     memset(&new_global_options, 0, sizeof(GLOBAL_OPTIONS)); /* reset global options */
@@ -1632,31 +1650,19 @@ int parse_conf(char *name, CONF_TYPE type) {
         if(config_opt[0]=='\0' || config_opt[0]=='#' || config_opt[0]==';') /* empty or comment */
             continue;
         if(config_opt[0]=='[' && config_opt[strlen(config_opt)-1]==']') { /* new section */
-            if(!section_init(line_number-1, section, 0)) {
+            if(section_init(line_number-1, section, 0)) {
                 file_close(df);
                 return 1;
             }
             ++config_opt;
             config_opt[strlen(config_opt)-1]='\0';
             new_section=str_alloc(sizeof(SERVICE_OPTIONS));
-            if(!new_section) {
-                s_log(LOG_ERR, "Fatal memory allocation error");
-                file_close(df);
-                return 1;
-            }
             memcpy(new_section, &new_service_options, sizeof(SERVICE_OPTIONS));
-            new_section->servname=str_dup_err(config_opt);
+            new_section->servname=str_dup(config_opt);
             new_section->session=NULL;
             new_section->next=NULL;
             section->next=new_section;
             section=new_section;
-#ifdef MAX_FD
-            if(++sections>MAX_FD) {
-                config_error(line_number, line_text, "Too many sections");
-                file_close(df);
-                return 1;
-            }
-#endif
             continue;
         }
         config_arg=strchr(config_line, '=');
@@ -1682,7 +1688,7 @@ int parse_conf(char *name, CONF_TYPE type) {
     file_close(df);
 
     /* initialize the last section */
-    if(!section_init(line_number, section, 1))
+    if(section_init(line_number, section, 1))
         return 1;
 
     /* final checks */
@@ -1697,38 +1703,65 @@ int parse_conf(char *name, CONF_TYPE type) {
             return 1;
         }
     }
-    memcpy(&service_options, &new_service_options, sizeof(SERVICE_OPTIONS));
     s_log(LOG_NOTICE, "Configuration successful");
     return 0;
 }
 
+void apply_conf() { /* can be used once the configuration was validated */
+    /* FIXME: this operation may be unsafe, as client() threads use it */
+    memcpy(&global_options, &new_global_options, sizeof(GLOBAL_OPTIONS));
+    /* service_options are used for inetd mode and to enumerate services */
+    memcpy(&service_options, &new_service_options, sizeof(SERVICE_OPTIONS));
+#ifdef USE_WIN32
+    PostMessage(hwnd, WM_VALID_CONFIG, 0, 0);
+#endif
+}
+
 /**************************************** validate and initialize section */
 
 static int section_init(int last_line, SERVICE_OPTIONS *section, int final) {
+    char *tmpstr;
+
+    /* setup host_name for SNI, prefer protocolHost if specified */
+    if(section->protocol_host) /* 'protocolHost' option */
+        section->host_name=str_dup(section->protocol_host);
+    else if(section->connect_name) /* 'connect' option */
+        section->host_name=str_dup(section->connect_name);
+    else
+        section->host_name=NULL;
+    if(section->host_name) { /* either 'protocolHost' or 'connect' specified */
+        tmpstr=strrchr(section->host_name, ':');
+        if(tmpstr) { /* 'host:port' -> drop ':port' */
+            *tmpstr='\0';
+        } else { /* 'port' -> default to 'localhost' */
+            str_free(section->host_name);
+            section->host_name=str_dup("localhost");
+        }
+    }
+
     if(section==&new_service_options) { /* global options just configured */
-        memcpy(&global_options, &new_global_options, sizeof(GLOBAL_OPTIONS));
 #ifdef HAVE_OSSL_ENGINE_H
         close_engine();
 #endif
-        if(!ssl_configure()) /* configure global SSL settings */
-            return 0;
+        if(ssl_configure(&new_global_options)) /* configure global SSL settings */
+            return 1;
         if(!final) /* no need to validate defaults */
-            return 1; /* OK */
+            return 0; /* OK */
     }
 
     if(!section->option.client && !section->cert) {
         section_error(last_line, section->servname,
             "SSL server needs a certificate");
-        return 0;
+        return 1;
     }
     if(context_init(section)) /* initialize SSL context */
-        return 0;
+        return 1;
 
     if(section==&new_service_options) { /* inetd mode checks */
         if(section->option.accept) {
             section_error(last_line, section->servname,
                 "'accept' is not allowed in inetd mode");
-            return 0;
+            return 1;
         }
 #if 0
         /* TODO: some additional checks could be useful */
@@ -1741,22 +1774,26 @@ static int section_init(int last_line, SERVICE_OPTIONS *section, int final) {
         if((unsigned int)section->option.accept
                 + (unsigned int)section->option.program
                 + (unsigned int)section->option.remote
+#ifndef OPENSSL_NO_TLSEXT
                 + (unsigned int)section->option.sni
+#endif /* OPENSSL_NO_TLSEXT */
 #ifndef USE_WIN32
                 + (unsigned int)section->option.transparent_dst
 #endif /* USE_WIN32 */
                 !=2) {
             section_error(last_line, section->servname,
                 "Each service must define two endpoints");
-            return 0;
+            return 1;
         }
+#ifndef OPENSSL_NO_TLSEXT
         if(section->option.sni && section->option.client) {
             section_error(last_line, section->servname,
                 "SNI slave service is a TLS client");
-            return 0;
+            return 1;
         }
+#endif /* OPENSSL_NO_TLSEXT */
     }
-    return 1; /* all tests passed -- continue program execution */
+    return 0; /* all tests passed -- continue program execution */
 }
 
 /**************************************** facility/debug level */
@@ -1803,7 +1840,7 @@ static int parse_debug_level(char *arg) {
         {NULL, -1}
     };
 
-    arg_copy=str_dup_err(arg);
+    arg_copy=str_dup(arg);
     string=arg_copy;
 
 /* facilities only make sense on Unix */
@@ -1819,7 +1856,7 @@ static int parse_debug_level(char *arg) {
             }
         }
         if(new_global_options.facility==-1)
-            return 0; /* FAILED */
+            return 1; /* FAILED */
         string=strtok(NULL, ".");    /* set to the remainder */
     }
 #endif /* USE_WIN32, __vms */
@@ -1827,7 +1864,7 @@ static int parse_debug_level(char *arg) {
     /* time to check the syslog level */
     if(string && strlen(string)==1 && *string>='0' && *string<='7') {
         new_global_options.debug_level=*string-'0';
-        return 1; /* OK */
+        return 0; /* OK */
     }
     new_global_options.debug_level=8;    /* illegal level */
     for(fl=levels; fl->name; ++fl) {
@@ -1837,8 +1874,8 @@ static int parse_debug_level(char *arg) {
         }
     }
     if(new_global_options.debug_level==8)
-        return 0; /* FAILED */
-    return 1; /* OK */
+        return 1; /* FAILED */
+    return 0; /* OK */
 }
 
 /**************************************** SSL options */
@@ -1850,6 +1887,9 @@ static int parse_ssl_option(char *arg) {
     } ssl_opts[] = {
         {"MICROSOFT_SESS_ID_BUG", SSL_OP_MICROSOFT_SESS_ID_BUG},
         {"NETSCAPE_CHALLENGE_BUG", SSL_OP_NETSCAPE_CHALLENGE_BUG},
+#ifdef SSL_OP_LEGACY_SERVER_CONNECT
+        {"LEGACY_SERVER_CONNECT", SSL_OP_LEGACY_SERVER_CONNECT},
+#endif
         {"NETSCAPE_REUSE_CIPHER_CHANGE_BUG",
             SSL_OP_NETSCAPE_REUSE_CIPHER_CHANGE_BUG},
         {"SSLREF2_REUSE_CERT_TYPE_BUG", SSL_OP_SSLREF2_REUSE_CERT_TYPE_BUG},
@@ -1858,7 +1898,10 @@ static int parse_ssl_option(char *arg) {
         {"SSLEAY_080_CLIENT_DH_BUG", SSL_OP_SSLEAY_080_CLIENT_DH_BUG},
         {"TLS_D5_BUG", SSL_OP_TLS_D5_BUG},
         {"TLS_BLOCK_PADDING_BUG", SSL_OP_TLS_BLOCK_PADDING_BUG},
+#ifdef SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
         {"DONT_INSERT_EMPTY_FRAGMENTS", SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS},
+#endif
+        {"ALL", SSL_OP_ALL},
 #ifdef SSL_OP_NO_QUERY_MTU
         {"NO_QUERY_MTU", SSL_OP_NO_QUERY_MTU},
 #endif
@@ -1868,17 +1911,28 @@ static int parse_ssl_option(char *arg) {
 #ifdef SSL_OP_NO_TICKET
         {"NO_TICKET", SSL_OP_NO_TICKET},
 #endif
+#ifdef SSL_OP_CISCO_ANYCONNECT
+        {"CISCO_ANYCONNECT", SSL_OP_CISCO_ANYCONNECT},
+#endif
+#ifdef SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION
         {"NO_SESSION_RESUMPTION_ON_RENEGOTIATION",
             SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION},
+#endif
 #ifdef SSL_OP_NO_COMPRESSION
         {"NO_COMPRESSION", SSL_OP_NO_COMPRESSION},
 #endif
+#ifdef SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
+        {"ALLOW_UNSAFE_LEGACY_RENEGOTIATION",
+            SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION},
+#endif
 #ifdef SSL_OP_SINGLE_ECDH_USE
         {"SINGLE_ECDH_USE", SSL_OP_SINGLE_ECDH_USE},
 #endif
         {"SINGLE_DH_USE", SSL_OP_SINGLE_DH_USE},
         {"EPHEMERAL_RSA", SSL_OP_EPHEMERAL_RSA},
+#ifdef SSL_OP_CIPHER_SERVER_PREFERENCE
         {"CIPHER_SERVER_PREFERENCE", SSL_OP_CIPHER_SERVER_PREFERENCE},
+#endif
         {"TLS_ROLLBACK_BUG", SSL_OP_TLS_ROLLBACK_BUG},
         {"NO_SSLv2", SSL_OP_NO_SSLv2},
         {"NO_SSLv3", SSL_OP_NO_SSLv3},
@@ -1891,7 +1945,6 @@ static int parse_ssl_option(char *arg) {
 #ifdef SSL_OP_CRYPTOPRO_TLSEXT_BUG
         {"CRYPTOPRO_TLSEXT_BUG", SSL_OP_CRYPTOPRO_TLSEXT_BUG},
 #endif
-        {"ALL", SSL_OP_ALL},
         {NULL, 0}
     }, *option;
 
@@ -1961,6 +2014,7 @@ static int print_socket_options(void) {
 
     fd=socket(AF_INET, SOCK_STREAM, 0);
 
+    s_log(LOG_NOTICE, " ");
     s_log(LOG_NOTICE, "Socket option defaults:");
     s_log(LOG_NOTICE,
         "    Option Name     |  Accept  |   Local  |  Remote  |OS default");
@@ -1971,10 +2025,10 @@ static int print_socket_options(void) {
         optlen=sizeof val;
         if(getsockopt(fd, ptr->opt_level,
                 ptr->opt_name, (void *)&val, &optlen)) {
-            if(get_last_socket_error()!=ENOPROTOOPT) {
+            if(get_last_socket_error()!=S_ENOPROTOOPT) {
                 s_log(LOG_ERR, "Failed to get %s OS default", ptr->opt_str);
                 sockerror("getsockopt");
-                return 0; /* FAILED */
+                return 1; /* FAILED */
             }
             td=str_dup("write-only");
         } else
@@ -1988,7 +2042,7 @@ static int print_socket_options(void) {
             ptr->opt_str, ta, tl, tr, td);
         str_free(ta); str_free(tl); str_free(tr); str_free(td);
     }
-    return 1; /* OK */
+    return 0; /* OK */
 }
 
 static char *print_option(int type, OPT_UNION *val) {
@@ -2017,7 +2071,7 @@ static int parse_socket_option(char *arg) {
     SOCK_OPT *ptr;
 
     if(arg[1]!=':')
-        return 0; /* FAILED */
+        return 1; /* FAILED */
     switch(arg[0]) {
     case 'a':
         socket_type=0; break;
@@ -2026,17 +2080,17 @@ static int parse_socket_option(char *arg) {
     case 'r':
         socket_type=2; break;
     default:
-        return 0; /* FAILED */
+        return 1; /* FAILED */
     }
     arg+=2;
     opt_val_str=strchr(arg, '=');
     if(!opt_val_str) /* no '='? */
-        return 0; /* FAILED */
+        return 1; /* FAILED */
     *opt_val_str++='\0';
     ptr=sock_opts;
     for(;;) {
         if(!ptr->opt_str)
-            return 0; /* FAILED */
+            return 1; /* FAILED */
         if(!strcmp(arg, ptr->opt_str))
             break; /* option name found */
         ++ptr;
@@ -2046,18 +2100,18 @@ static int parse_socket_option(char *arg) {
     case TYPE_FLAG:
         if(!strcasecmp(opt_val_str, "yes") || !strcmp(opt_val_str, "1")) {
             ptr->opt_val[socket_type]->i_val=1;
-            return 1; /* OK */
+            return 0; /* OK */
         }
         if(!strcasecmp(opt_val_str, "no") || !strcmp(opt_val_str, "0")) {
             ptr->opt_val[socket_type]->i_val=0;
-            return 1; /* OK */
+            return 0; /* OK */
         }
-        return 0; /* FAILED */
+        return 1; /* FAILED */
     case TYPE_INT:
         ptr->opt_val[socket_type]->i_val=strtol(opt_val_str, &tmpstr, 10);
         if(tmpstr==arg || *tmpstr) /* not a number */
-            return 0; /* FAILED */
-        return 1; /* OK */
+            return 1; /* FAILED */
+        return 0; /* OK */
     case TYPE_LINGER:
         opt_val2_str=strchr(opt_val_str, ':');
         if(opt_val2_str) {
@@ -2065,42 +2119,44 @@ static int parse_socket_option(char *arg) {
             ptr->opt_val[socket_type]->linger_val.l_linger=
                 (u_short)strtol(opt_val2_str, &tmpstr, 10);
             if(tmpstr==arg || *tmpstr) /* not a number */
-                return 0; /* FAILED */
+                return 1; /* FAILED */
         } else {
             ptr->opt_val[socket_type]->linger_val.l_linger=0;
         }
         ptr->opt_val[socket_type]->linger_val.l_onoff=
             (u_short)strtol(opt_val_str, &tmpstr, 10);
         if(tmpstr==arg || *tmpstr) /* not a number */
-            return 0; /* FAILED */
-        return 1; /* OK */
+            return 1; /* FAILED */
+        return 0; /* OK */
     case TYPE_TIMEVAL:
         opt_val2_str=strchr(opt_val_str, ':');
         if(opt_val2_str) {
             *opt_val2_str++='\0';
             ptr->opt_val[socket_type]->timeval_val.tv_usec=strtol(opt_val2_str, &tmpstr, 10);
             if(tmpstr==arg || *tmpstr) /* not a number */
-                return 0; /* FAILED */
+                return 1; /* FAILED */
         } else {
             ptr->opt_val[socket_type]->timeval_val.tv_usec=0;
         }
         ptr->opt_val[socket_type]->timeval_val.tv_sec=strtol(opt_val_str, &tmpstr, 10);
         if(tmpstr==arg || *tmpstr) /* not a number */
-            return 0; /* FAILED */
-        return 1; /* OK */
+            return 1; /* FAILED */
+        return 0; /* OK */
     case TYPE_STRING:
         if(strlen(opt_val_str)+1>sizeof(OPT_UNION))
-            return 0; /* FAILED */
+            return 1; /* FAILED */
         strcpy(ptr->opt_val[socket_type]->c_val, opt_val_str);
-        return 1; /* OK */
+        return 0; /* OK */
     default:
         ; /* ANSI C compiler needs it */
     }
-    return 0; /* FAILED */
+    return 1; /* FAILED */
 }
 
 /**************************************** OCSP */
 
+#ifdef HAVE_OSSL_OCSP_H
+
 static char *parse_ocsp_url(SERVICE_OPTIONS *section, char *arg) {
     char *host, *port, *path;
     int ssl;
@@ -2110,9 +2166,9 @@ static char *parse_ocsp_url(SERVICE_OPTIONS *section, char *arg) {
     if(ssl)
         return "SSL not supported for OCSP"
             " - additional stunnel service needs to be defined";
-    if(!hostport2addrlist(&section->ocsp_addr, host, port))
+    if(!hostport2addr(&section->ocsp_addr, host, port))
         return "Failed to resolve OCSP server address";
-    section->ocsp_path=str_dup_err(path);
+    section->ocsp_path=str_dup(path);
     if(host)
         OPENSSL_free(host);
     if(port)
@@ -2148,6 +2204,91 @@ static unsigned long parse_ocsp_flag(char *arg) {
     return 0; /* FAILED */
 }
 
+#endif /* HAVE_OSSL_OCSP_H */
+
+/**************************************** engine */
+
+#ifdef HAVE_OSSL_ENGINE_H
+
+#define MAX_ENGINES 256
+static ENGINE *engines[MAX_ENGINES]; /* table of engines */
+static int current_engine=0;
+static int engine_initialized;
+
+static char *open_engine(const char *name) {
+    s_log(LOG_DEBUG, "Enabling support for engine '%s'", name);
+    if(!strcasecmp(name, "auto")) {
+        ENGINE_register_all_complete();
+        s_log(LOG_DEBUG, "Auto engine support enabled");
+        return NULL; /* OK */
+    }
+
+    close_engine(); /* close the previous one (if specified) */
+    engines[current_engine]=ENGINE_by_id(name);
+    engine_initialized=0;
+    if(!engines[current_engine]) {
+        sslerror("ENGINE_by_id");
+        return "Failed to open the engine";
+    }
+    return NULL; /* OK */
+}
+
+static char *ctrl_engine(const char *cmd, const char *arg) {
+    if(!strcasecmp(cmd, "INIT")) { /* special control command */
+        return init_engine();
+    }
+    if(arg)
+        s_log(LOG_DEBUG, "Executing engine control command %s:%s", cmd, arg);
+    else
+        s_log(LOG_DEBUG, "Executing engine control command %s", cmd);
+    if(!ENGINE_ctrl_cmd_string(engines[current_engine], cmd, arg, 0)) {
+        sslerror("ENGINE_ctrl_cmd_string");
+        return "Failed to execute the engine control command";
+    }
+    return NULL; /* OK */
+}
+
+static char *init_engine(void) {
+    if(engine_initialized)
+        return NULL; /* OK */
+    engine_initialized=1;
+    s_log(LOG_DEBUG, "Initializing engine %d", current_engine+1);
+    if(!ENGINE_init(engines[current_engine])) {
+        if(ERR_peek_last_error()) /* really an error */
+            sslerror("ENGINE_init");
+        else
+            s_log(LOG_ERR, "Engine %d not initialized", current_engine+1);
+        return "Engine initialization failed";
+    }
+    if(!ENGINE_set_default(engines[current_engine], ENGINE_METHOD_ALL)) {
+        sslerror("ENGINE_set_default");
+        return "Selecting default engine failed";
+    }
+    s_log(LOG_DEBUG, "Engine %d initialized", current_engine+1);
+    return NULL; /* OK */
+}
+
+static void close_engine(void) {
+    if(!engines[current_engine])
+        return; /* no engine was opened -> nothing to do */
+    init_engine();
+    ++current_engine;
+#if 0
+    ENGINE_finish(e);
+    ENGINE_free(e);
+    e=NULL;
+    s_log(LOG_DEBUG, "Engine closed");
+#endif
+}
+
+static ENGINE *get_engine(int i) {
+    if(i<1 || i>current_engine)
+        return NULL;
+    return engines[i-1];
+}
+
+#endif /* HAVE_OSSL_ENGINE_H */
+
 /**************************************** fatal error */
 
 static void print_syntax(void) {
@@ -2180,7 +2321,7 @@ static void print_syntax(void) {
     s_log(LOG_NOTICE, "    -sockets    - display default socket options");
 }
 
-/**************************************** various supporting funstions */
+/**************************************** various supporting functions */
 
 static void config_error(int num, const char *line, const char *str) {
     s_log(LOG_ERR, "Line %d: \"%s\": %s", num, line, str);
@@ -2190,17 +2331,6 @@ static void section_error(int num, const char *name, const char *str) {
     s_log(LOG_ERR, "Line %d: End of section %s: %s", num, name, str);
 }
 
-static char *str_dup_err(char *str) { /* str_dup() with error checking */
-    char *retval;
-
-    retval=str_dup(str);
-    if(!retval) {
-        s_log(LOG_ERR, "Fatal memory allocation error");
-        die(2);
-    }
-    return retval;
-}
-
 #ifndef USE_WIN32
 
 static char **argalloc(char *str) { /* allocate 'exec' argumets */
@@ -2208,12 +2338,8 @@ static char **argalloc(char *str) { /* allocate 'exec' argumets */
     char *ptr, **retval;
 
     max_arg=strlen(str)/2+1;
-    ptr=str_dup_err(str);
+    ptr=str_dup(str);
     retval=str_alloc((max_arg+1)*sizeof(char *));
-    if(!retval) {
-        s_log(LOG_ERR, "Fatal memory allocation error");
-        die(2);
-    }
     i=0;
     while(*ptr && i<max_arg) {
         retval[i++]=ptr;
