@@ -38,11 +38,10 @@
 #include "common.h"
 #include "prototypes.h"
 
-/* \n is not a character expected in the string */
-#define LINE "%[^\n]"
 #define isprefix(a, b) (strncasecmp((a), (b), strlen(b))==0)
 
 /* protocol-specific function prototypes */
+static void proxy_server(CLI *c);
 static void cifs_client(CLI *);
 static void cifs_server(CLI *);
 static void pgsql_client(CLI *);
@@ -54,62 +53,122 @@ static void pop3_server(CLI *);
 static void imap_client(CLI *);
 static void imap_server(CLI *);
 static void nntp_client(CLI *);
+static void connect_server(CLI *);
 static void connect_client(CLI *);
+
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
 static void ntlm(CLI *);
-#ifndef OPENSSL_NO_MD4
 static char *ntlm1();
 static char *ntlm3(char *, char *, char *);
 static void crypt_DES(DES_cblock, DES_cblock, DES_cblock);
 #endif
 static char *base64(int, char *, int);
 
-void negotiate(CLI *c) {
-    if(!c->opt->protocol)
-        return; /* no protocol negotiations */
+/**************************************** framework */
 
-    s_log(LOG_NOTICE, "Negotiations for %s (%s side) started", c->opt->protocol,
-        c->opt->option.client ? "client" : "server");
+typedef void (*FUNCTION)(CLI *);
 
-    if(c->opt->option.client) {
-        if(!strcmp(c->opt->protocol, "cifs"))
-            cifs_client(c);
-        else if(!strcmp(c->opt->protocol, "smtp"))
-            smtp_client(c);
-        else if(!strcmp(c->opt->protocol, "pop3"))
-            pop3_client(c);
-        else if(!strcmp(c->opt->protocol, "imap"))
-            imap_client(c);
-        else if(!strcmp(c->opt->protocol, "nntp"))
-            nntp_client(c);
-        else if(!strcmp(c->opt->protocol, "connect"))
-            connect_client(c);
-        else if(!strcmp(c->opt->protocol, "pgsql"))
-            pgsql_client(c);
-        else {
-            s_log(LOG_ERR, "Protocol %s not supported in client mode",
-                c->opt->protocol);
-            longjmp(c->err, 1);
-        }
-    } else {
-        if(!strcmp(c->opt->protocol, "cifs"))
-            cifs_server(c);
-        else if(!strcmp(c->opt->protocol, "smtp"))
-            smtp_server(c);
-        else if(!strcmp(c->opt->protocol, "pop3"))
-            pop3_server(c);
-        else if(!strcmp(c->opt->protocol, "imap"))
-            imap_server(c);
-        else if(!strcmp(c->opt->protocol, "pgsql"))
-            pgsql_server(c);
-        else {
-            s_log(LOG_ERR, "Protocol %s not supported in server mode",
-                c->opt->protocol);
-            longjmp(c->err, 1);
-        }
-    }
-    s_log(LOG_NOTICE, "Protocol negotiations succeeded");
+static const struct {
+    char *name;
+    struct {
+        PROTOCOL_TYPE type;
+        FUNCTION func;
+    } handlers[2];
+} protocols[]={
+    {"proxy",   {{PROTOCOL_PRE_SSL,     proxy_server},      {PROTOCOL_PRE_SSL, NULL}}},
+    {"cifs",    {{PROTOCOL_PRE_CONNECT, cifs_server},       {PROTOCOL_PRE_SSL, cifs_client}}},
+    {"pgsql",   {{PROTOCOL_PRE_CONNECT, pgsql_server},      {PROTOCOL_PRE_SSL, pgsql_client}}},
+    {"smtp",    {{PROTOCOL_PRE_SSL,     smtp_server},       {PROTOCOL_PRE_SSL, smtp_client}}},
+    {"pop3",    {{PROTOCOL_PRE_SSL,     pop3_server},       {PROTOCOL_PRE_SSL, pop3_client}}},
+    {"imap",    {{PROTOCOL_PRE_SSL,     imap_server},       {PROTOCOL_PRE_SSL, imap_client}}},
+    {"nntp",    {{PROTOCOL_NONE,        NULL},              {PROTOCOL_PRE_SSL, nntp_client}}},
+    {"connect", {{PROTOCOL_PRE_CONNECT, connect_server},    {PROTOCOL_PRE_SSL, connect_client}}},
+    {NULL,      {{PROTOCOL_NONE,        NULL},              {PROTOCOL_NONE,    NULL}}}
+};
+
+int find_protocol_id(const char *name) {
+    int id;
+
+    for(id=0; protocols[id].name; ++id)
+        if(!strcmp(name, protocols[id].name))
+            return id;
+    return -1;
 }
 
+void protocol(CLI *c, const PROTOCOL_TYPE type) {
+    const int id=c->opt->protocol, mode=(unsigned int)c->opt->option.client;
+
+    if(id<0 || type!=protocols[id].handlers[mode].type ||
+            !protocols[id].handlers[mode].func)
+        return;
+    s_log(LOG_INFO, "%s-mode %s protocol negotiations started",
+        mode ? "Client" : "Server", protocols[id].name);
+    protocols[id].handlers[mode].func(c);
+    s_log(LOG_INFO, "%s-mode %s protocol negotiations succeeded",
+        mode ? "Client" : "Server", protocols[id].name);
+}
+
+/**************************************** proxy */
+
+/*
+ * PROXY protocol: http://haproxy.1wt.eu/download/1.5/doc/proxy-protocol.txt
+ * this is a protocol client support for stunnel acting as an SSL server 
+ * I don't think anything else is useful, but feel free to discuss on the
+ * stunnel-users mailing list if you disagree
+ */
+
+/* IP address textual representation length */
+/* 1234:6789:1234:6789:1234:6789:1234:6789 -> 40 chars with '\0' */
+#define IP_LEN 40
+#define PORT_LEN 6
+
+static void proxy_server(CLI *c) {
+    SOCKADDR_UNION addr;
+    socklen_t addrlen;
+    char src_host[IP_LEN], dst_host[IP_LEN];
+    char src_port[PORT_LEN], dst_port[PORT_LEN], *proto;
+    int err;
+
+    addrlen=sizeof addr;
+    if(getpeername(c->local_rfd.fd, &addr.sa, &addrlen)) {
+        sockerror("getpeername");
+        longjmp(c->err, 1);
+    }
+    err=getnameinfo(&addr.sa, addr_len(&addr), src_host, IP_LEN,
+        src_port, PORT_LEN, NI_NUMERICHOST|NI_NUMERICSERV);
+    if(err) {
+        s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
+        longjmp(c->err, 1);
+    }
+
+    addrlen=sizeof addr;
+    if(getsockname(c->local_rfd.fd, &addr.sa, &addrlen)) {
+        sockerror("getsockname");
+        longjmp(c->err, 1);
+    }
+    err=getnameinfo(&addr.sa, addr_len(&addr), dst_host, IP_LEN,
+        dst_port, PORT_LEN, NI_NUMERICHOST|NI_NUMERICSERV);
+    if(err) {
+        s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
+        longjmp(c->err, 1);
+    }
+
+    switch(addr.sa.sa_family) {
+    case AF_INET:
+        proto="TCP4";
+        break;
+    case AF_INET6:
+        proto="TCP6";
+        break;
+    default: /* AF_UNIX */
+        proto="UNKNOWN";
+    }
+    fd_printf(c, c->remote_fd.fd, "PROXY %s %s %s %s %s",
+        proto, src_host, dst_host, src_port, dst_port);
+}
+
+/**************************************** cifs */
+
 static void cifs_client(CLI *c) {
     u8 buffer[5];
     u8 request_dummy[4] = {0x81, 0, 0, 0}; /* a zero-length request */
@@ -152,6 +211,8 @@ static void cifs_server(CLI *c) {
     write_blocking(c, c->local_wfd.fd, response_use_ssl, 5);
 }
 
+/**************************************** pgsql */
+
 /* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73982 */
 u8 ssl_request[8]={0, 0, 0, 8, 0x04, 0xd2, 0x16, 0x2f};
 
@@ -180,26 +241,28 @@ static void pgsql_server(CLI *c) {
     write_blocking(c, c->local_wfd.fd, ssl_ok, sizeof ssl_ok);
 }
 
+/**************************************** smtp */
+
 static void smtp_client(CLI *c) {
     char *line;
 
     do { /* copy multiline greeting */
-        line=fdgetline(c, c->remote_fd.fd);
-        fdputline(c, c->local_wfd.fd, line);
+        line=fd_getline(c, c->remote_fd.fd);
+        fd_putline(c, c->local_wfd.fd, line);
     } while(isprefix(line, "220-"));
 
-    fdputline(c, c->remote_fd.fd, "EHLO localhost");
+    fd_putline(c, c->remote_fd.fd, "EHLO localhost");
     do { /* skip multiline reply */
-        line=fdgetline(c, c->remote_fd.fd);
+        line=fd_getline(c, c->remote_fd.fd);
     } while(isprefix(line, "250-"));
     if(!isprefix(line, "250 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 1425 compliant");
         longjmp(c->err, 1);
     }
 
-    fdputline(c, c->remote_fd.fd, "STARTTLS");
+    fd_putline(c, c->remote_fd.fd, "STARTTLS");
     do { /* skip multiline reply */
-        line=fdgetline(c, c->remote_fd.fd);
+        line=fd_getline(c, c->remote_fd.fd);
     } while(isprefix(line, "220-"));
     if(!isprefix(line, "220 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 2487 compliant");
@@ -210,9 +273,9 @@ static void smtp_client(CLI *c) {
 static void smtp_server(CLI *c) {
     char *line;
 
-    s_poll_init(&c->fds);
-    s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
-    switch(s_poll_wait(&c->fds, 0, 200)) { /* wait up to 200ms */
+    s_poll_init(c->fds);
+    s_poll_add(c->fds, c->local_rfd.fd, 1, 0);
+    switch(s_poll_wait(c->fds, 0, 200)) { /* wait up to 200ms */
     case 0: /* fd not ready to read */
         s_log(LOG_DEBUG, "RFC 2487 detected");
         break;
@@ -224,38 +287,40 @@ static void smtp_server(CLI *c) {
         longjmp(c->err, 1);
     }
 
-    line=fdgetline(c, c->remote_fd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "220")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
-    fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    line=fdgetline(c, c->local_rfd.fd);
+    fd_printf(c, c->local_wfd.fd, "%s + stunnel", line);
+    line=fd_getline(c, c->local_rfd.fd);
     if(!isprefix(line, "EHLO ")) {
         s_log(LOG_ERR, "Unknown client EHLO");
         longjmp(c->err, 1);
     }
-    fdprintf(c, c->local_wfd.fd, "250-%s Welcome", line);
-    fdputline(c, c->local_wfd.fd, "250 STARTTLS");
-    line=fdgetline(c, c->local_rfd.fd);
+    fd_printf(c, c->local_wfd.fd, "250-%s Welcome", line);
+    fd_putline(c, c->local_wfd.fd, "250 STARTTLS");
+    line=fd_getline(c, c->local_rfd.fd);
     if(!isprefix(line, "STARTTLS")) {
         s_log(LOG_ERR, "STARTTLS expected");
         longjmp(c->err, 1);
     }
-    fdputline(c, c->local_wfd.fd, "220 Go ahead");
+    fd_putline(c, c->local_wfd.fd, "220 Go ahead");
 }
 
+/**************************************** pop3 */
+
 static void pop3_client(CLI *c) {
     char *line;
 
-    line=fdgetline(c, c->remote_fd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
-    fdputline(c, c->local_wfd.fd, line);
-    fdputline(c, c->remote_fd.fd, "STLS");
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->local_wfd.fd, line);
+    fd_putline(c, c->remote_fd.fd, "STLS");
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
@@ -265,33 +330,37 @@ static void pop3_client(CLI *c) {
 static void pop3_server(CLI *c) {
     char *line;
 
-    line=fdgetline(c, c->remote_fd.fd);
-    fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    line=fdgetline(c, c->local_rfd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
+    fd_printf(c, c->local_wfd.fd, "%s + stunnel", line);
+    line=fd_getline(c, c->local_rfd.fd);
     if(isprefix(line, "CAPA")) { /* client wants RFC 2449 extensions */
-        fdputline(c, c->local_wfd.fd, "-ERR Stunnel does not support capabilities");
-        line=fdgetline(c, c->local_rfd.fd);
+        fd_putline(c, c->local_wfd.fd, "+OK Stunnel capability list follows");
+        fd_putline(c, c->local_wfd.fd, "STLS");
+        fd_putline(c, c->local_wfd.fd, ".");
+        line=fd_getline(c, c->local_rfd.fd);
     }
     if(!isprefix(line, "STLS")) {
         s_log(LOG_ERR, "Client does not want TLS");
         longjmp(c->err, 1);
     }
-    fdputline(c, c->local_wfd.fd, "+OK Stunnel starts TLS negotiation");
+    fd_putline(c, c->local_wfd.fd, "+OK Stunnel starts TLS negotiation");
 }
 
+/**************************************** imap */
+
 static void imap_client(CLI *c) {
     char *line;
 
-    line=fdgetline(c, c->remote_fd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
-    fdputline(c, c->local_wfd.fd, line);
-    fdputline(c, c->remote_fd.fd, "stunnel STARTTLS");
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->local_wfd.fd, line);
+    fd_putline(c, c->remote_fd.fd, "stunnel STARTTLS");
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "stunnel OK")) {
-        fdputline(c, c->local_wfd.fd,
+        fd_putline(c, c->local_wfd.fd,
             "* BYE stunnel: Server does not support TLS");
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 2); /* don't reset */
@@ -301,9 +370,9 @@ static void imap_client(CLI *c) {
 static void imap_server(CLI *c) {
     char *line, *id, *tail, *capa;
  
-    s_poll_init(&c->fds);
-    s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
-    switch(s_poll_wait(&c->fds, 0, 200)) {
+    s_poll_init(c->fds);
+    s_poll_add(c->fds, c->local_rfd.fd, 1, 0);
+    switch(s_poll_wait(c->fds, 0, 200)) {
     case 0: /* fd not ready to read */
         s_log(LOG_DEBUG, "RFC 2595 detected");
         break;
@@ -316,7 +385,7 @@ static void imap_server(CLI *c) {
     }
 
     /* process server welcome and send it to client */
-    line=fdgetline(c, c->remote_fd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
@@ -326,10 +395,10 @@ static void imap_server(CLI *c) {
         capa=strstr(line, "capability");
     if(capa)
         *capa='K'; /* disable CAPABILITY within greeting */
-    fdprintf(c, c->local_wfd.fd, "%s (stunnel)", line);
+    fd_printf(c, c->local_wfd.fd, "%s (stunnel)", line);
 
     while(1) { /* process client commands */
-        line=fdgetline(c, c->local_rfd.fd);
+        line=fd_getline(c, c->local_rfd.fd);
         /* split line into id and tail */
         id=str_dup(line);
         tail=strchr(id, ' ');
@@ -338,12 +407,12 @@ static void imap_server(CLI *c) {
         *tail++='\0';
 
         if(isprefix(tail, "STARTTLS")) {
-            fdprintf(c, c->local_wfd.fd,
+            fd_printf(c, c->local_wfd.fd,
                 "%s OK Begin TLS negotiation now", id);
             return; /* success */
         } else if(isprefix(tail, "CAPABILITY")) {
-            fdputline(c, c->remote_fd.fd, line); /* send it to server */
-            line=fdgetline(c, c->remote_fd.fd); /* get the capabilites */
+            fd_putline(c, c->remote_fd.fd, line); /* send it to server */
+            line=fd_getline(c, c->remote_fd.fd); /* get the capabilites */
             if(*line=='*') {
                 /* 
                  * append STARTTLS
@@ -351,53 +420,93 @@ static void imap_server(CLI *c) {
                  * of Mozilla bug #324138/#312009
                  * LOGIN would fail as "unexpected command", anyway
                  */
-                fdprintf(c, c->local_wfd.fd, "%s STARTTLS", line);
-                line=fdgetline(c, c->remote_fd.fd); /* next line */
+                fd_printf(c, c->local_wfd.fd, "%s STARTTLS", line);
+                line=fd_getline(c, c->remote_fd.fd); /* next line */
             }
-            fdputline(c, c->local_wfd.fd, line); /* forward to the client */
+            fd_putline(c, c->local_wfd.fd, line); /* forward to the client */
             tail=strchr(line, ' ');
             if(!tail || !isprefix(tail+1, "OK")) { /* not OK? */
-                fdputline(c, c->local_wfd.fd,
+                fd_putline(c, c->local_wfd.fd,
                     "* BYE unexpected server response");
                 s_log(LOG_ERR, "Unexpected server response: %s", line);
                 break;
             }
         } else if(isprefix(tail, "LOGOUT")) {
-            fdputline(c, c->local_wfd.fd, "* BYE server terminating");
-            fdprintf(c, c->local_wfd.fd, "%s OK LOGOUT completed", id);
+            fd_putline(c, c->local_wfd.fd, "* BYE server terminating");
+            fd_printf(c, c->local_wfd.fd, "%s OK LOGOUT completed", id);
             break;
         } else {
-            fdputline(c, c->local_wfd.fd, "* BYE stunnel: unexpected command");
-            fdprintf(c, c->local_wfd.fd, "%s BAD %s unexpected", id, tail);
+            fd_putline(c, c->local_wfd.fd, "* BYE stunnel: unexpected command");
+            fd_printf(c, c->local_wfd.fd, "%s BAD %s unexpected", id, tail);
             s_log(LOG_ERR, "Unexpected client command %s", tail);
             break;
         }
     }    
     /* clean server shutdown */
-    fdputline(c, c->remote_fd.fd, "stunnel LOGOUT");
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->remote_fd.fd, "stunnel LOGOUT");
+    line=fd_getline(c, c->remote_fd.fd);
     if(*line=='*')
-        line=fdgetline(c, c->remote_fd.fd);
+        line=fd_getline(c, c->remote_fd.fd);
     longjmp(c->err, 2); /* don't reset */
 }
 
+/**************************************** nntp */
+
 static void nntp_client(CLI *c) {
     char *line;
 
-    line=fdgetline(c, c->remote_fd.fd);
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "200 ") && !isprefix(line, "201 ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
-    fdputline(c, c->local_wfd.fd, line);
-    fdputline(c, c->remote_fd.fd, "STARTTLS");
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->local_wfd.fd, line);
+    fd_putline(c, c->remote_fd.fd, "STARTTLS");
+    line=fd_getline(c, c->remote_fd.fd);
     if(!isprefix(line, "382 ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
     }
 }
 
+/**************************************** connect */
+
+static void connect_server(CLI *c) {
+    char *request, *proto, *header;
+    int not_empty;
+
+    request=fd_getline(c, c->local_rfd.fd);
+    if(!isprefix(request, "CONNECT ")) {
+        fd_putline(c, c->local_wfd.fd, "HTTP/1.0 400 Bad Request Method");
+        fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
+        fd_putline(c, c->local_wfd.fd, "");
+        longjmp(c->err, 1);
+    }
+    proto=strchr(request+8, ' ');
+    if(!proto || !isprefix(proto, " HTTP/")) {
+        fd_putline(c, c->local_wfd.fd, "HTTP/1.0 400 Bad Request Protocol");
+        fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
+        fd_putline(c, c->local_wfd.fd, "");
+        longjmp(c->err, 1);
+    }
+    *proto='\0';
+    do { /* ignore any headers*/
+        header=fd_getline(c, c->local_rfd.fd);
+        not_empty=*header;
+        str_free(header);
+    } while(not_empty);
+    if(!name2addrlist(&c->connect_addr, request+8, DEFAULT_LOOPBACK)) {
+        fd_putline(c, c->local_wfd.fd, "HTTP/1.0 404 Not Found");
+        fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
+        fd_putline(c, c->local_wfd.fd, "");
+        longjmp(c->err, 1);
+    }
+    str_free(request);
+    fd_putline(c, c->local_wfd.fd, "HTTP/1.0 200 OK");
+    fd_putline(c, c->local_wfd.fd, "Server: stunnel/" STUNNEL_VERSION);
+    fd_putline(c, c->local_wfd.fd, "");
+}
+
 static void connect_client(CLI *c) {
     char *line, *encoded;
 
@@ -405,12 +514,17 @@ static void connect_client(CLI *c) {
         s_log(LOG_ERR, "protocolHost not specified");
         longjmp(c->err, 1);
     }
-    fdprintf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1",
+    fd_printf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1",
         c->opt->protocol_host);
-    fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
+    fd_printf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
     if(c->opt->protocol_username && c->opt->protocol_password) {
         if(!strcasecmp(c->opt->protocol_authentication, "NTLM")) {
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
             ntlm(c);
+#else
+            s_log(LOG_ERR, "NTLM authentication is not available");
+            longjmp(c->err, 1);
+#endif
         } else { /* basic authentication */
             line=str_printf("%s:%s",
                 c->opt->protocol_username, c->opt->protocol_password);
@@ -420,27 +534,29 @@ static void connect_client(CLI *c) {
                 s_log(LOG_ERR, "Base64 encoder failed");
                 longjmp(c->err, 1);
             }
-            fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: basic %s",
+            fd_printf(c, c->remote_fd.fd, "Proxy-Authorization: basic %s",
                 encoded);
             str_free(encoded);
         }
     }
-    fdputline(c, c->remote_fd.fd, ""); /* empty line */
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->remote_fd.fd, ""); /* empty line */
+    line=fd_getline(c, c->remote_fd.fd);
     if(strlen(line)<12 || line[9]!='2') {
         /* not "HTTP/1.0 200 Connection established" */
         s_log(LOG_ERR, "CONNECT request rejected");
         do { /* read all headers */
-            line=fdgetline(c, c->remote_fd.fd);
+            line=fd_getline(c, c->remote_fd.fd);
         } while(*line);
         longjmp(c->err, 1);
     }
     s_log(LOG_INFO, "CONNECT request accepted");
     do {
-        line=fdgetline(c, c->remote_fd.fd); /* read all headers */
+        line=fd_getline(c, c->remote_fd.fd); /* read all headers */
     } while(*line);
 }
 
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
+
 /* 
  * NTLM code is based on the following documentation:
  * http://davenport.sourceforge.net/ntlm.html
@@ -450,33 +566,32 @@ static void connect_client(CLI *c) {
 #define s_min(a, b) ((a)>(b)?(b):(a))
 
 static void ntlm(CLI *c) {
-#ifndef OPENSSL_NO_MD4
     char *line, buf[BUFSIZ], *ntlm1_txt, *ntlm2_txt, *ntlm3_txt;
     long content_length=0; /* no HTTP content */
 
     /* send Proxy-Authorization (phase 1) */
-    fdprintf(c, c->remote_fd.fd, "Proxy-Connection: keep-alive");
+    fd_printf(c, c->remote_fd.fd, "Proxy-Connection: keep-alive");
     ntlm1_txt=ntlm1();
     if(!ntlm1_txt) {
         s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM request");
         longjmp(c->err, 1);
     }
-    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm1_txt);
+    fd_printf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm1_txt);
     str_free(ntlm1_txt);
-    fdputline(c, c->remote_fd.fd, ""); /* empty line */
-    line=fdgetline(c, c->remote_fd.fd);
+    fd_putline(c, c->remote_fd.fd, ""); /* empty line */
+    line=fd_getline(c, c->remote_fd.fd);
 
     /* receive Proxy-Authenticate (phase 2) */
     if(line[9]!='4' || line[10]!='0' || line[11]!='7') { /* code 407 */
         s_log(LOG_ERR, "NTLM authorization request rejected");
         do { /* read all headers */
-            line=fdgetline(c, c->remote_fd.fd);
+            line=fd_getline(c, c->remote_fd.fd);
         } while(*line);
         longjmp(c->err, 1);
     }
     ntlm2_txt=NULL;
     do { /* read all headers */
-        line=fdgetline(c, c->remote_fd.fd);
+        line=fd_getline(c, c->remote_fd.fd);
         if(isprefix(line, "Proxy-Authenticate: NTLM "))
             ntlm2_txt=str_dup(line+25);
         else if(isprefix(line, "Content-Length: "))
@@ -494,24 +609,18 @@ static void ntlm(CLI *c) {
     }
 
     /* send Proxy-Authorization (phase 3) */
-    fdprintf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1", c->opt->protocol_host);
-    fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
+    fd_printf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1", c->opt->protocol_host);
+    fd_printf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
     ntlm3_txt=ntlm3(c->opt->protocol_username, c->opt->protocol_password, ntlm2_txt);
     str_free(ntlm2_txt);
     if(!ntlm3_txt) {
         s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM response");
         longjmp(c->err, 1);
     }
-    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm3_txt);
+    fd_printf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm3_txt);
     str_free(ntlm3_txt);
-#else
-    s_log(LOG_ERR, "NTLM authentication is not available");
-    longjmp(c->err, 1);
-#endif
 }
 
-#ifndef OPENSSL_NO_MD4
-
 static char *ntlm1() {
     char phase1[16];
 
@@ -625,11 +734,6 @@ static char *base64(int encode, char *in, int len) {
     /* 32 bytes as a safety precaution for passing decoded data to crypt_DES */
     /* n+1 to get null-terminated string on encode */
     out=str_alloc(n<32?32:n+1);
-    if(!out) {
-        s_log(LOG_ERR, "Memory allocation error");
-        BIO_free_all(bio);
-        return NULL;
-    }
     n=BIO_read(bio, out, n);
     if(n<0) {
         BIO_free_all(bio);
