@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -38,7 +38,7 @@
 #include "common.h"
 #include "prototypes.h"
 
-/**************************************** Resolver functions */
+/**************************************** prototypes */
 
 #ifndef HAVE_GETADDRINFO
 
@@ -52,12 +52,11 @@
 #define EAI_SERVICE 8
 #endif
 
-#if defined(USE_WIN32) && defined(USE_IPv6)
+#ifdef USE_WIN32
 /* rename some locally shadowed declarations */
 #define getaddrinfo     local_getaddrinfo
 #define freeaddrinfo    local_freeaddrinfo
-#define getnameinfo     local_getnameinfo
-#else /* defined(USE_WIN32) && defined(USE_IPv6) */
+#else /* defined USE_WIN32 */
 struct addrinfo {
     int ai_flags;
     int ai_family;
@@ -68,7 +67,7 @@ struct addrinfo {
     char *ai_canonname;
     struct addrinfo *ai_next;
 };
-#endif /* defined(USE_WIN32) && defined(USE_IPv6) */
+#endif /* defined USE_WIN32 */
 
 static int getaddrinfo(const char *, const char *,
     const struct addrinfo *, struct addrinfo **);
@@ -78,24 +77,60 @@ static void freeaddrinfo(struct addrinfo *);
 
 #endif /* !defined HAVE_GETADDRINFO */
 
-static const char *s_gai_strerror(int);
+/**************************************** stunnel resolver API */
 
-#ifndef HAVE_GETNAMEINFO
-#ifndef NI_NUMERICHOST
-#define NI_NUMERICHOST  2
-#endif
-#ifndef NI_NUMERICSERV
-#define NI_NUMERICSERV  8
-#endif
-static int getnameinfo(const struct sockaddr *, int,
-    char *, int , char *, int , int );
-#endif
+int name2addr(SOCKADDR_UNION *addr, char *name, char *default_host) {
+    SOCKADDR_LIST addr_list;
+    int retval;
+
+    addr_list.num=0;
+    addr_list.addr=NULL;
+    retval=name2addrlist(&addr_list, name, default_host);
+    if(retval>0)
+        memcpy(addr, &addr_list.addr[0], sizeof *addr);
+    if(addr_list.addr)
+        str_free(addr_list.addr);
+    return retval;
+}
+
+int hostport2addr(SOCKADDR_UNION *addr, char *hostname, char *portname) {
+    SOCKADDR_LIST addr_list;
+    int retval;
+
+    addr_list.num=0;
+    addr_list.addr=NULL;
+    retval=hostport2addrlist(&addr_list, hostname, portname);
+    if(retval>0)
+        memcpy(addr, &addr_list.addr[0], sizeof *addr);
+    if(addr_list.addr)
+        str_free(addr_list.addr);
+    return retval;
+}
 
 int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
-    char tmp[STRLEN], *hostname, *portname;
+    char *tmp, *hostname, *portname;
+    int retval;
+
+    addr_list->cur=0; /* reset round-robin counter */
+
+    /* first check if this is a UNIX socket */
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    if(*name=='/') {
+        if(offsetof(struct sockaddr_un, sun_path)+strlen(name)+1
+                > sizeof(struct sockaddr_un)) {
+            s_log(LOG_ERR, "Unix socket path is too long");
+            return 0; /* no results */
+        }
+        addr_list->addr=str_realloc(addr_list->addr,
+            (addr_list->num+1)*sizeof(SOCKADDR_UNION));
+        addr_list->addr[addr_list->num].un.sun_family=AF_UNIX;
+        strcpy(addr_list->addr[addr_list->num].un.sun_path, name);
+        return ++(addr_list->num); /* ok - return the number of addresses */
+    }
+#endif
 
     /* set hostname and portname */
-    safecopy(tmp, name);
+    tmp=str_dup(name);
     portname=strrchr(tmp, ':');
     if(portname) {
         hostname=tmp;
@@ -106,7 +141,9 @@ int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
     }
 
     /* fill addr_list structure */
-    return hostport2addrlist(addr_list, hostname, portname);
+    retval=hostport2addrlist(addr_list, hostname, portname);
+    str_free(tmp);
+    return retval;
 }
 
 int hostport2addrlist(SOCKADDR_LIST *addr_list,
@@ -114,8 +151,6 @@ int hostport2addrlist(SOCKADDR_LIST *addr_list,
     struct addrinfo hints, *res=NULL, *cur;
     int err;
 
-    addr_list->cur=0; /* initialize round-robin counter */
-
     memset(&hints, 0, sizeof hints);
 #if defined(USE_IPv6) || defined(USE_WIN32)
     hints.ai_family=PF_UNSPEC;
@@ -124,52 +159,81 @@ int hostport2addrlist(SOCKADDR_LIST *addr_list,
 #endif
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
-
-    err=getaddrinfo(hostname, portname, &hints, &res);
-    if(err) {
-        if(err==EAI_SERVICE)
-            s_log(LOG_ERR, "Unknown TCP service '%s'", portname);
-        else
-            s_log(LOG_ERR, "Error resolving '%s': %s",
-                hostname, s_gai_strerror(err));
-        if(res)
+    do {
+        err=getaddrinfo(hostname, portname, &hints, &res);
+        if(err && res)
             freeaddrinfo(res);
-        return 0; /* Error */
+        if(err==EAI_AGAIN) {
+            s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
+            sleep(1);
+        }
+    } while(err==EAI_AGAIN);
+    switch(err) {
+    case 0:
+        break; /* success */
+    case EAI_SERVICE:
+        s_log(LOG_ERR, "Unknown TCP service '%s'", portname);
+        return 0; /* error */
+    default:
+        s_log(LOG_ERR, "Error resolving '%s': %s",
+            hostname, s_gai_strerror(err));
+        return 0; /* error */
     }
+
     /* copy the list of addresses */
-    cur=res;
-    while(cur && addr_list->num<MAX_HOSTS) {
-        if(cur->ai_addrlen>sizeof(SOCKADDR_UNION)) {
+    for(cur=res; cur; cur=cur->ai_next) {
+        if(cur->ai_addrlen>(int)sizeof(SOCKADDR_UNION)) {
             s_log(LOG_ERR, "INTERNAL ERROR: ai_addrlen value too big");
             freeaddrinfo(res);
             return 0; /* no results */
         }
-        memcpy(&addr_list->addr[addr_list->num],
-            cur->ai_addr, cur->ai_addrlen);
-        cur=cur->ai_next;
-        addr_list->num++;
+        addr_list->addr=str_realloc(addr_list->addr,
+            (addr_list->num+1)*sizeof(SOCKADDR_UNION));
+        memcpy(&addr_list->addr[addr_list->num], cur->ai_addr, cur->ai_addrlen);
+        ++(addr_list->num);
     }
     freeaddrinfo(res);
     return addr_list->num; /* ok - return the number of addresses */
 }
 
-char *s_ntop(char *text, SOCKADDR_UNION *addr) {
-    char host[IPLEN-6], port[6];
+char *s_ntop(SOCKADDR_UNION *addr, socklen_t addrlen) {
+    int err;
+    char *host, *port, *retval;
 
-    if(getnameinfo(&addr->sa, addr_len(*addr),
-            host, IPLEN-6, port, 6, NI_NUMERICHOST|NI_NUMERICSERV)) {
-        sockerror("getnameinfo");
-        strcpy(text, "unresolvable IP");
-        return text;
-    }
-    strcpy(text, host);
-    strcat(text, ":");
-    strcat(text, port);
-    return text;
+    if(addrlen==sizeof(u_short)) /* see UNIX(7) manual for details */
+        return str_dup("unnamed socket");
+    host=str_alloc(256);
+    port=str_alloc(256); /* needs to be long enough for AF_UNIX path */
+    err=getnameinfo(&addr->sa, addrlen,
+        host, 256, port, 256, NI_NUMERICHOST|NI_NUMERICSERV);
+    if(err) {
+        s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
+        retval=str_dup("unresolvable address");
+    } else
+        retval=str_printf("%s:%s", host, port);
+    str_free(host);
+    str_free(port);
+    return retval;
 }
 
-/**************************************** My getaddrinfo() and getnameinfo() */
-/* implementations are limited to functionality needed by stunnel */
+socklen_t addr_len(const SOCKADDR_UNION *addr) {
+    if(addr->sa.sa_family==AF_INET)
+        return sizeof(struct sockaddr_in);
+#ifdef USE_IPv6
+    if(addr->sa.sa_family==AF_INET6)
+        return sizeof(struct sockaddr_in6);
+#endif
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    if(addr->sa.sa_family==AF_UNIX)
+        return sizeof(struct sockaddr_un);
+#endif
+    s_log(LOG_ERR, "INTERNAL ERROR: Unknown sa_family: %d",
+        addr->sa.sa_family);
+    return sizeof(SOCKADDR_UNION);
+}
+
+/**************************************** my getaddrinfo() */
+/* implementation is limited to functionality needed by stunnel */
 
 #ifndef HAVE_GETADDRINFO
 static int getaddrinfo(const char *node, const char *service,
@@ -181,14 +245,15 @@ static int getaddrinfo(const char *node, const char *service,
     u_short port;
     struct addrinfo *ai;
     int retval;
+    char *tmpstr;
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
     if(s_getaddrinfo)
         return s_getaddrinfo(node, service, hints, res);
 #endif
     /* decode service name */
-    port=htons((u_short)atoi(service));
-    if(!port) { /* zero is an illegal value for port number */
+    port=htons((u_short)strtol(service, &tmpstr, 10));
+    if(tmpstr==service || *tmpstr) { /* not a number */
 #ifdef _WIN32_WCE
         return EAI_NONAME;
 #else /* defined(_WIN32_WCE) */
@@ -200,9 +265,7 @@ static int getaddrinfo(const char *node, const char *service,
     }
 
     /* allocate addrlist structure */
-    ai=calloc(1, sizeof(struct addrinfo));
-    if(!ai)
-        return EAI_MEMORY;
+    ai=str_alloc(sizeof(struct addrinfo));
     if(hints)
         memcpy(ai, hints, sizeof(struct addrinfo));
 
@@ -210,22 +273,14 @@ static int getaddrinfo(const char *node, const char *service,
 #if defined(USE_IPv6) && !defined(USE_WIN32)
     ai->ai_family=AF_INET6;
     ai->ai_addrlen=sizeof(struct sockaddr_in6);
-    ai->ai_addr=calloc(1, ai->ai_addrlen);
-    if(!ai->ai_addr) {
-        free(ai);
-        return EAI_MEMORY;
-    }
+    ai->ai_addr=str_alloc(ai->ai_addrlen);
     ai->ai_addr->sa_family=AF_INET6;
     if(inet_pton(AF_INET6, node,
             &((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr)>0) {
 #else
     ai->ai_family=AF_INET;
     ai->ai_addrlen=sizeof(struct sockaddr_in);
-    ai->ai_addr=calloc(1, ai->ai_addrlen);
-    if(!ai->ai_addr) {
-        free(ai);
-        return EAI_MEMORY;
-    }
+    ai->ai_addr=str_alloc(ai->ai_addrlen);
     ai->ai_addr->sa_family=AF_INET;
     ((struct sockaddr_in *)ai->ai_addr)->sin_addr.s_addr=inet_addr(node);
     if(((struct sockaddr_in *)ai->ai_addr)->sin_addr.s_addr+1) {
@@ -235,8 +290,8 @@ static int getaddrinfo(const char *node, const char *service,
         *res=ai;
         return 0; /* numerical address resolved */
     }
-    free(ai->ai_addr);
-    free(ai);
+    str_free(ai->ai_addr);
+    str_free(ai);
 
     /* not numerical: need to call resolver library */
     *res=NULL;
@@ -274,9 +329,7 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
 
     /* copy addresses */
     for(i=0; h->h_addr_list[i]; i++) {
-        ai=calloc(1, sizeof(struct addrinfo));
-        if(!ai)
-            return EAI_MEMORY;
+        ai=str_alloc(sizeof(struct addrinfo));
         if(hints)
             memcpy(ai, hints, sizeof(struct addrinfo));
         ai->ai_next=NULL; /* just in case */
@@ -291,18 +344,14 @@ static int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
 #if defined(USE_IPv6)
         if(h->h_addrtype==AF_INET6) {
             ai->ai_addrlen=sizeof(struct sockaddr_in6);
-            ai->ai_addr=calloc(1, ai->ai_addrlen);
-            if(!ai->ai_addr)
-                return EAI_MEMORY;
+            ai->ai_addr=str_alloc(ai->ai_addrlen);
             memcpy(&((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr,
                 h->h_addr_list[i], h->h_length);
         } else
 #endif
         {
             ai->ai_addrlen=sizeof(struct sockaddr_in);
-            ai->ai_addr=calloc(1, ai->ai_addrlen);
-            if(!ai->ai_addr)
-                return EAI_MEMORY;
+            ai->ai_addr=str_alloc(ai->ai_addrlen);
             memcpy(&((struct sockaddr_in *)ai->ai_addr)->sin_addr,
                 h->h_addr_list[i], h->h_length);
         }
@@ -324,19 +373,18 @@ static void freeaddrinfo(struct addrinfo *current) {
 #endif
     while(current) {
         if(current->ai_addr)
-            free(current->ai_addr);
+            str_free(current->ai_addr);
         if(current->ai_canonname)
-            free(current->ai_canonname);
+            str_free(current->ai_canonname);
         next=current->ai_next;
-        free(current);
+        str_free(current);
         current=next;
     }
 }
-
 #endif /* !defined HAVE_GETADDRINFO */
 
-/* Due to a problem with Mingw32 I decided to define my own gai_strerror() */
-static const char *s_gai_strerror(int err) {
+/* due to a problem with Mingw32 I decided to define my own gai_strerror() */
+const char *s_gai_strerror(int err) {
     switch(err) {
 #ifdef EAI_BADFLAGS
         case EAI_BADFLAGS:
@@ -385,8 +433,11 @@ static const char *s_gai_strerror(int err) {
     }
 }
 
+/**************************************** my getnameinfo() */
+/* implementation is limited to functionality needed by stunnel */
+
 #ifndef HAVE_GETNAMEINFO
-static int getnameinfo(const struct sockaddr *sa, int salen,
+int getnameinfo(const struct sockaddr *sa, int salen,
     char *host, int hostlen, char *serv, int servlen, int flags) {
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
@@ -415,4 +466,4 @@ static int getnameinfo(const struct sockaddr *sa, int salen,
 }
 #endif
 
-/* End of resolver.c */
+/* end of resolver.c */
