@@ -46,6 +46,9 @@
 #endif /* __va_copy */
 #endif /* va_copy */
 
+static u8 canary[10]; /* 80-bit canary value */
+static volatile int canary_initialized=0;
+
 typedef struct alloc_list ALLOC_LIST;
 
 typedef struct alloc_tls {
@@ -57,19 +60,19 @@ struct alloc_list {
     ALLOC_LIST *prev, *next;
     ALLOC_TLS *tls;
     size_t size;
+    int valid_canary;
     unsigned int magic;
 };
 
 static void set_alloc_tls(ALLOC_TLS *);
 static ALLOC_TLS *get_alloc_tls();
-static ALLOC_LIST *get_alloc_list_ptr(void *);
+static ALLOC_LIST *get_alloc_list_ptr(void *, char *, int);
 
 char *str_dup(const char *str) {
     char *retval;
 
     retval=str_alloc(strlen(str)+1);
-    if(retval)
-        strcpy(retval, str);
+    strcpy(retval, str);
     return retval;
 }
 
@@ -89,8 +92,6 @@ char *str_vprintf(const char *format, va_list start_ap) {
     va_list ap;
 
     p=str_alloc(size);
-    if(!p)
-        return NULL;
     for(;;) {
         va_copy(ap, start_ap);
         n=vsnprintf(p, size, format, ap);
@@ -101,50 +102,48 @@ char *str_vprintf(const char *format, va_list start_ap) {
         else          /* glibc 2.0, WIN32, etc. */
             size*=2;  /* twice the old size */
         np=str_realloc(p, size);
-        if(!np) {
-            str_free(p);
-            return NULL;
-        }
         p=np; /* LOL */
     }
 }
 
 #ifdef USE_UCONTEXT
 
-static ALLOC_TLS *alloc_tls=NULL;
+static ALLOC_TLS *global_tls=NULL;
 
 void str_init() {
 }
 
-static void set_alloc_tls(ALLOC_TLS *alloc_tls) {
+static void set_alloc_tls(ALLOC_TLS *tls) {
     if(ready_head)
-        ready_head->tls=alloc_tls;
+        ready_head->tls=tls;
     else /* ucontext threads not initialized */
-        alloc_tls=alloc_tls;
+        global_tls=tls;
 }
 
 static ALLOC_TLS *get_alloc_tls() {
     if(ready_head)
         return ready_head->tls;
     else /* ucontext threads not initialized */
-        return alloc_tls;
+        return global_tls;
 }
 
 #endif /* USE_UCONTEXT */
 
 #ifdef USE_FORK
 
-static ALLOC_TLS *alloc_tls=NULL;
+static ALLOC_TLS *global_tls;
+static pid_t previous_pid=0;
 
 void str_init() {
 }
 
-static void set_alloc_tls(ALLOC_TLS *alloc_tls) {
-    alloc_tls=alloc_tls;
+static void set_alloc_tls(ALLOC_TLS *tls) {
+    global_tls=tls;
+    previous_pid=getpid();
 }
 
 static ALLOC_TLS *get_alloc_tls() {
-    return alloc_tls;
+    return previous_pid==getpid() ? global_tls : NULL;
 }
 
 #endif /* USE_FORK */
@@ -157,8 +156,8 @@ void str_init() {
     pthread_key_create(&pthread_key, NULL);
 }
 
-static void set_alloc_tls(ALLOC_TLS *alloc_tls) {
-    pthread_setspecific(pthread_key, alloc_tls);
+static void set_alloc_tls(ALLOC_TLS *tls) {
+    pthread_setspecific(pthread_key, tls);
 }
 
 static ALLOC_TLS *get_alloc_tls() {
@@ -173,92 +172,94 @@ static DWORD tls_index;
 
 void str_init() {
     tls_index=TlsAlloc();
-    if(tls_index==TLS_OUT_OF_INDEXES) {
-        s_log(LOG_ERR, "TlsAlloc failed");
-        die(1);
-    }
 }
 
 static void set_alloc_tls(ALLOC_TLS *alloc_tls) {
-    if(!TlsSetValue(tls_index, alloc_tls)) {
-        s_log(LOG_ERR, "TlsSetValue failed");
-        die(1);
-    }
+    TlsSetValue(tls_index, alloc_tls);
 }
 
 static ALLOC_TLS *get_alloc_tls() {
-    ALLOC_TLS *alloc_tls;
-
-    alloc_tls=TlsGetValue(tls_index);
-    if(!alloc_tls && GetLastError()!=ERROR_SUCCESS) {
-        s_log(LOG_ERR, "TlsGetValue failed");
-        die(1);
-    }
-    return alloc_tls;
+    return TlsGetValue(tls_index);
 }
 
 #endif /* USE_WIN32 */
 
+void str_canary_init() {
+    RAND_bytes(canary, sizeof canary);
+    canary_initialized=1; /* after RAND_bytes */
+}
+
 void str_cleanup() {
     ALLOC_TLS *alloc_tls;
 
     alloc_tls=get_alloc_tls();
-    while(alloc_tls->head)
-        str_free(alloc_tls->head+1);
-    set_alloc_tls(NULL);
-    free(alloc_tls);
+    if(alloc_tls) {
+        while(alloc_tls->head)
+            str_free(alloc_tls->head+1);
+        set_alloc_tls(NULL);
+        free(alloc_tls);
+    }
 }
 
 void str_stats() {
     ALLOC_TLS *alloc_tls;
 
     alloc_tls=get_alloc_tls();
-    s_log(LOG_DEBUG, "str_stats: %d block(s), %d byte(s)",
-        alloc_tls->blocks, alloc_tls->bytes);
+    if(alloc_tls)
+        s_log(LOG_DEBUG,
+            "str_stats: %lu block(s), %lu data byte(s), %lu control byte(s)",
+            (unsigned long int)alloc_tls->blocks,
+            (unsigned long int)alloc_tls->bytes,
+            (unsigned long int)(alloc_tls->blocks*
+                (sizeof(ALLOC_LIST)+sizeof canary)));
+    else
+        s_log(LOG_DEBUG, "str_stats: alloc_tls not initialized");
 }
 
-void *str_alloc(size_t size) {
+void *str_alloc_debug(size_t size, char *file, int line) {
     ALLOC_TLS *alloc_tls;
     ALLOC_LIST *alloc_list;
 
-    if(size>=1024*1024) /* huge allocations are not allowed */
-        return NULL;
     alloc_tls=get_alloc_tls();
     if(!alloc_tls) { /* first allocation in this thread */
         alloc_tls=calloc(1, sizeof(ALLOC_TLS));
         if(!alloc_tls)
-            return NULL;
+            fatal("Out of memory", file, line);
         alloc_tls->head=NULL;
         alloc_tls->bytes=alloc_tls->blocks=0;
         set_alloc_tls(alloc_tls);
     }
-    alloc_list=calloc(1, sizeof(ALLOC_LIST)+size);
+    alloc_list=calloc(1, sizeof(ALLOC_LIST)+size+sizeof canary);
     if(!alloc_list)
-        return NULL;
+        fatal("Out of memory", file, line);
+
     alloc_list->prev=NULL;
     alloc_list->next=alloc_tls->head;
     alloc_list->tls=alloc_tls;
     alloc_list->size=size;
+    alloc_list->valid_canary=canary_initialized; /* before memcpy */
+    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
     alloc_list->magic=0xdeadbeef;
+
     if(alloc_tls->head)
         alloc_tls->head->prev=alloc_list;
     alloc_tls->head=alloc_list;
     alloc_tls->bytes+=size;
     alloc_tls->blocks++;
+
     return alloc_list+1;
 }
 
-void *str_realloc(void *ptr, size_t size) {
+void *str_realloc_debug(void *ptr, size_t size, char *file, int line) {
     ALLOC_LIST *previous_alloc_list, *alloc_list;
 
     if(!ptr)
         return str_alloc(size);
-    if(size>=1024*1024) /* huge allocations are not allowed */
-        return NULL;
-    previous_alloc_list=get_alloc_list_ptr(ptr);
-    alloc_list=realloc(previous_alloc_list, sizeof(ALLOC_LIST)+size);
+    previous_alloc_list=get_alloc_list_ptr(ptr, file, line);
+    alloc_list=realloc(previous_alloc_list,
+        sizeof(ALLOC_LIST)+size+sizeof canary);
     if(!alloc_list)
-        return NULL;
+        fatal("Out of memory", file, line);
     if(alloc_list->tls) { /* not detached */
         /* refresh possibly invalidated linked list pointers */
         if(alloc_list->tls->head==previous_alloc_list)
@@ -271,17 +272,19 @@ void *str_realloc(void *ptr, size_t size) {
         alloc_list->tls->bytes+=size-alloc_list->size;
     }
     alloc_list->size=size;
+    alloc_list->valid_canary=canary_initialized; /* before memcpy */
+    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
     return alloc_list+1;
 }
 
 /* detach from thread automatic deallocation list */
 /* it has no effect if the allocation is already detached */
-void str_detach(void *ptr) {
+void str_detach_debug(void *ptr, char *file, int line) {
     ALLOC_LIST *alloc_list;
 
     if(!ptr) /* do not attempt to free null pointers */
         return;
-    alloc_list=get_alloc_list_ptr(ptr);
+    alloc_list=get_alloc_list_ptr(ptr, file, line);
     if(alloc_list->tls) { /* not detached */
         /* remove from linked list */
         if(alloc_list->tls->head==alloc_list)
@@ -300,29 +303,28 @@ void str_detach(void *ptr) {
     }
 }
 
-void str_free(void *ptr) {
+void str_free_debug(void *ptr, char *file, int line) {
     ALLOC_LIST *alloc_list;
 
     if(!ptr) /* do not attempt to free null pointers */
         return;
-    str_detach(ptr);
+    str_detach_debug(ptr, file, line);
     alloc_list=(ALLOC_LIST *)ptr-1;
     alloc_list->magic=0xdefec8ed; /* to detect double free */
     free(alloc_list);
 }
 
-static ALLOC_LIST *get_alloc_list_ptr(void *ptr) {
+static ALLOC_LIST *get_alloc_list_ptr(void *ptr, char *file, int line) {
     ALLOC_LIST *alloc_list;
 
     alloc_list=(ALLOC_LIST *)ptr-1;
-    if(alloc_list->magic!=0xdeadbeef) { /* not allocated by str_alloc() */
-        s_log(LOG_CRIT, "INTERNAL ERROR: get_alloc_list_ptr: Bad magic");
-        die(1);
-    }
-    if(alloc_list->tls /* not detached */ && alloc_list->tls!=get_alloc_tls()) {
-        s_log(LOG_CRIT, "INTERNAL ERROR: get_alloc_list_ptr: Wrong thread");
-        die(1);
-    }
+    if(alloc_list->magic!=0xdeadbeef) /* not allocated by str_alloc() */
+        fatal("Bad magic", file, line);
+    if(alloc_list->tls /* not detached */ && alloc_list->tls!=get_alloc_tls())
+        fatal("Wrong thread", file, line);
+    if(alloc_list->valid_canary &&
+            memcmp((u8 *)ptr+alloc_list->size, canary, sizeof canary))
+        fatal("Dead canary", file, line);
     return alloc_list;
 }
 
