@@ -48,7 +48,9 @@ static int add_dir_lookup(X509_STORE *, char *);
 static int verify_callback(int, X509_STORE_CTX *);
 static int cert_check(CLI *c, X509_STORE_CTX *, int);
 static int crl_check(CLI *c, X509_STORE_CTX *);
+#ifdef HAVE_OSSL_OCSP_H
 static int ocsp_check(CLI *c, X509_STORE_CTX *);
+#endif
 
 /* utility functions */
 static void log_time(const int, const char *, ASN1_TIME *);
@@ -183,12 +185,14 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
         OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
+#ifdef HAVE_OSSL_OCSP_H
     if(c->opt->option.ocsp && !ocsp_check(c, callback_ctx)) {
         s_log(LOG_WARNING, "OCSP check failed: depth=%d, %s",
             callback_ctx->error_depth, subject_name);
         OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
+#endif /* HAVE_OSSL_OCSP_H */
     /* errnum=X509_STORE_CTX_get_error(ctx); */
     s_log(LOG_NOTICE, "Certificate accepted: depth=%d, %s",
         callback_ctx->error_depth, subject_name);
@@ -200,7 +204,9 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
 
 static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
     X509_OBJECT obj;
+#if OPENSSL_VERSION_NUMBER>=0x0090700fL
     ASN1_BIT_STRING *local_key, *peer_key;
+#endif
 
     if(c->opt->verify_level<1) {
         s_log(LOG_INFO, "CERT: Verification not enabled");
@@ -208,9 +214,14 @@ static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
     }
     if(!preverify_ok) {
         /* remote site specified a certificate, but it's not correct */
-        s_log(LOG_WARNING, "CERT: Verification error: %s",
-            X509_verify_cert_error_string(callback_ctx->error));
-        return 0; /* reject connection */
+        if(c->opt->verify_level>=4 && callback_ctx->error_depth>0) {
+            s_log(LOG_INFO, "CERT: Invalid CA certificate ignored");
+            return 1; /* accept connection */
+        } else {    
+            s_log(LOG_WARNING, "CERT: Verification error: %s",
+                X509_verify_cert_error_string(callback_ctx->error));
+            return 0; /* reject connection */
+        }
     }
     if(c->opt->verify_level>=3 && callback_ctx->error_depth==0) {
         if(X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
@@ -218,6 +229,7 @@ static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
             s_log(LOG_WARNING, "CERT: Certificate not found in local repository");
             return 0; /* reject connection */
         }
+#if OPENSSL_VERSION_NUMBER>=0x0090700fL
         peer_key=X509_get0_pubkey_bitstr(callback_ctx->current_cert);
         local_key=X509_get0_pubkey_bitstr(obj.data.x509);
         if(!peer_key || !local_key || peer_key->length!=local_key->length ||
@@ -225,6 +237,7 @@ static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
             s_log(LOG_WARNING, "CERT: Public keys do not match");
             return 0; /* reject connection */
         }
+#endif
         s_log(LOG_INFO, "CERT: Locally installed certificate matched");
     }
     return 1; /* accept connection */
@@ -328,12 +341,13 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     return 1; /* accept connection */
 }
 
+#ifdef HAVE_OSSL_OCSP_H
+
 /**************************************** OCSP checking */
 /* TODO: check OCSP server specified in the certificate */
 
 static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     int error, retval=0;
-    SOCKADDR_UNION addr;
     X509 *cert;
     X509 *issuer=NULL;
     OCSP_CERTID *certID;
@@ -346,12 +360,11 @@ static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     int status, reason;
 
     /* connect specified OCSP server (responder) */
-    c->fd=s_socket(c->opt->ocsp_addr.addr[0].sa.sa_family, SOCK_STREAM, 0,
+    c->fd=s_socket(c->opt->ocsp_addr.sa.sa_family, SOCK_STREAM, 0,
         0, "OCSP: socket (auth_user)");
     if(c->fd<0)
         return 0; /* reject connection */
-    memcpy(&addr, &c->opt->ocsp_addr.addr[0], sizeof addr);
-    if(connect_blocking(c, &addr, addr_len(addr)))
+    if(connect_blocking(c, &c->opt->ocsp_addr, addr_len(&c->opt->ocsp_addr)))
         goto cleanup;
     s_log(LOG_DEBUG, "OCSP: server connected");
 
@@ -451,6 +464,8 @@ cleanup:
     return retval;
 }
 
+#endif /* HAVE_OSSL_OCSP_H */
+
 static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     char *cp;
     BIO *bio;
@@ -464,10 +479,6 @@ static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     ASN1_TIME_print(bio, t);
     n=BIO_pending(bio);
     cp=str_alloc(n+1);
-    if(!cp) {
-        BIO_free(bio);
-        return;
-    }
     n=BIO_read(bio, cp, n);
     if(n<0) {
         BIO_free(bio);
